# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import numpy as np
import os
import opfython.math.general as g
import opfython.stream.parser as p
import opfython.stream.splitter as s
from opfython.models import SemiSupervisedOPF

from ..utils import import_mnist_dataset


def import_data():
    # import the dataset
    (X_train, Y_train), (X_test, Y_test) = import_mnist_dataset()

    # if needed perform some preprocessing
    SAMPLES_LIMIT = 5000
    X_train = X_train.astype('float32') / 255
    X_test = X_test.astype('float32') / 255
    X_train, Y_train = X_train[:int(
        0.7*SAMPLES_LIMIT)], Y_train[:int(0.7*SAMPLES_LIMIT)]
    X_test, Y_test = X_test[:int(0.3*SAMPLES_LIMIT)
                            ], Y_test[:int(0.3*SAMPLES_LIMIT)]
    return X_train, X_test, Y_train, Y_test


def import_classifier():
    # upload a classifier
    # !!! the classifier must be a tf.keras.models.Model !!!
    # !!! change the next line with the path to your classifier !!!
    classifier_path = os.path.join("tmp", "MNIST", "classifier")
    classifier = tf.keras.models.load_model(classifier_path)
    return classifier


def generate_classifier(X_train, Y_train):
    input_shape = X_train.shape[1:]
    num_classes = len(np.unique(Y_train))

    classifier = tf.keras.models.Sequential([
        tf.keras.layers.Flatten(input_shape=input_shape),
        tf.keras.layers.Dense(num_classes, activation=tf.nn.softmax),
    ])

    classifier.compile(
        optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
    classifier.fit(X_train, Y_train, epochs=20)
    return classifier


def opf(X_train, Y_train):
    pr_x = int(0.1*len(X_train))
    pr_y = int(0.1*len(Y_train))
    X_train, X_unlabeled, Y_train, Y_unlabeled = X_train[:
                                                         pr_x], X_train[pr_x:], Y_train[:pr_y], Y_train[pr_y:]

    # Creates a SemiSupervisedOPF instance
    opf = SemiSupervisedOPF(
        distance="log_squared_euclidean", pre_computed_distance=None)

    # Fits training data along with unlabeled data into the semi-supervised classifier
    opf.fit(X_train, Y_train, X_unlabeled)

    # Predicts new data
    preds = opf.predict(X_unlabeled)

    # Calculating accuracy
    acc = g.opf_accuracy(Y_unlabeled, preds)
    print(f"OPF Accuracy: {acc}")

    return np.hstack((Y_train, preds))


def import_2d_data(projection='t-SNE'):
    # upload the 2D projection of the data
    with open(os.path.join("tmp", "MNIST", "DBM", projection, "train_2d.npy"), "rb") as f:
        X2d_train = np.load(f)
    with open(os.path.join("tmp", "MNIST", "DBM", projection, "test_2d.npy"), "rb") as f:
        X2d_test = np.load(f)
    return X2d_train, X2d_test



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy as np

from .NNInv import DEFAULT_MODEL_PATH, NNInv
from .projections import PROJECTION_METHODS


from ..AbstractDBM import AbstractDBM, DBM_DEFAULT_RESOLUTION, DEFAULT_TRAINING_EPOCHS, DEFAULT_BATCH_SIZE, FAST_DBM_STRATEGIES

from ...utils import track_time_wrapper, TRAIN_DATA_POINT_MARKER, TEST_DATA_POINT_MARKER, TRAIN_2D_FILE_NAME, TEST_2D_FILE_NAME
from ...Logger import LoggerInterface, Logger
time_tracker_console = Logger(name="Decision Boundary Mapper - DBM", info_color="cyan", show_init=False)

CUSTOM_PROJECTION_NAME = "Custom"

class DBM(AbstractDBM):
    """
    This is a class that generated the DBM
    The core purpose of this class is to learn the inverse projection given a direct projection and construct the decision boundary map of a given classifier
    
    Public methods:
        fit: Learns an inverse projection by training a neural network. \n
        get_decision_boundary_map: Returns the decision boundary map for the given classifier. \n

    Example:
        >>> from DBM import DBM
        >>> import matplotlib.pyplot as plt
        >>> classifier = ...
        >>> dbm = DBM(classifier)
        >>> img, img_confidence, encoded_2d_train, encoded_2d_test = dbm.get_decision_boundary_map(Xnd_train, Xnd_test, X2d_train, X2d_test)
        >>> plt.imshow(img)
        >>> plt.show()
    """

    def __init__(self, classifier, logger: LoggerInterface | None = None):
        """ 
        Initializes the DBM class.

        Args:
            classifier (tensorflow.keras.Model): The classifier to be used for the decision boundary map.
            logger (LoggerInterface, optional): The logger for outputting info messages. Defaults to console logging.
        """
        super().__init__(classifier, logger)
        self.neural_network = None  # type: ignore

    @track_time_wrapper(logger=time_tracker_console)
    def fit(self,
            X2d: np.ndarray, Xnd: np.ndarray,
            epochs: int = DEFAULT_TRAINING_EPOCHS, batch_size: int = DEFAULT_BATCH_SIZE,
            load_folder: str = DEFAULT_MODEL_PATH):
        """ 
        Learns the inverse projection on the given data set.

        Args:
            X2d (np.ndarray): Training data set 2D data got from the projection of the original data (e.g. PCA, t-SNE, UMAP)
            Xnd (np.ndarray): Training data set nD data (e.g. MNIST, CIFAR10) (i.e. the original data)
            epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 300.
            batch_size (int, optional): Train batch size. Defaults to 32.

        Returns:
            inverse_porjection_NN (NNInv): The trained inverse projection neural network.
        """

        inverse_projection_NN = NNInv(folder_path=load_folder, logger=self.console)
        inverse_projection_NN.fit(X2d, Xnd,
                                  epochs=epochs,
                                  batch_size=batch_size)
        return inverse_projection_NN

    def generate_boundary_map(self,
                              Xnd_train: np.ndarray,
                              Xnd_test: np.ndarray,
                              X2d_train: np.ndarray | None = None,
                              X2d_test: np.ndarray | None = None,
                              nn_train_epochs: int = DEFAULT_TRAINING_EPOCHS,
                              nn_train_batch_size: int = DEFAULT_BATCH_SIZE,
                              resolution: int = DBM_DEFAULT_RESOLUTION,
                              fast_decoding_strategy: FAST_DBM_STRATEGIES = FAST_DBM_STRATEGIES.NONE,
                              load_folder: str = DEFAULT_MODEL_PATH,
                              projection: str = 't-SNE'):
        """ 
        Generates a 2D boundary map of the classifier's decision boundary.

        Args:
            X_train (np.ndarray): Training data set (nD)
            X_test (np.ndarray): Testing data set (nD)
            X2d_train (np.ndarray | None): The 2D projection of the training data. If None, the data is projected using the given projection method. Defaults to None.
            X2d_test (np.ndarray | None): The 2D projection of the testing data. If None, the data is projected using the given projection method. Defaults to None.            
            nn_train_epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 300.
            nn_train_batch_size (int, optional): Train batch size. Defaults to 32.
            resolution (int, optional): The desired resolution of the DBM. Defaults to DBM_DEFAULT_RESOLUTION.
            fast_decoding_strategy (FAST_DBM_STRATEGIES, optional): The strategy to use in generating the DBM. Defaults to FAST_DBM_STRATEGIES.NONE.
            load_folder (str, optional): The folder in which the model will be stored or if exists loaded from. Defaults to DEFAULT_MODEL_PATH
            projection (str, optional): The projection method to be used. Defaults to 't-SNE'.

        Returns:
            img (np.array): A 2D numpy array with the decision boundary map, each element is an integer representing the class of the corresponding point.
            img_confidence (np.array): A 2D numpy array with the decision boundary map, each element is a float representing the confidence of the classifier for the corresponding point.
            encoded_2d_train (np.array): An array representing the projection of the training data set, each element is a tuple representing the coordinates and the class of the corresponding point.
            encoded_2d_test (np.array): An array representing the projection of the testing data set, each element is an tuple representing the coordinates and the class of the corresponding point.
    
        Example:
            >>> dbm = DBM(classifier)
            >>> img, img_confidence, _, _ = dbm.generate_boundary_map(X_train, X_test)
            >>> fig, [ax1, ax2] = plt.subplots(1, 2, figsize=(10, 5))
            >>> ax1.imshow(img)
            >>> ax2.imshow(img_confidence)
            >>> plt.show()
        """
       
        # adding projection method to the end of the load_folder path
        if projection != load_folder.split(os.sep)[-1]:
            load_folder = os.path.join(load_folder, projection)
            
        if X2d_train is None or X2d_test is None:
            assert projection in PROJECTION_METHODS.keys()
            Xnd_train_flatten = Xnd_train.reshape((Xnd_train.shape[0], -1))
            Xnd_test_flatten = Xnd_test.reshape((Xnd_test.shape[0], -1))
            X2d_train, X2d_test = self.__transform_2d__(Xnd_train_flatten, Xnd_test_flatten, load_folder, projection)
        else:
            # Normalize the data to be in the range of [0,1]
            X2d_train, X2d_test = self.__normalize_2d__(X2d_train, X2d_test)
            

        # creating a folder for the model if not present
        if not os.path.exists(os.path.join(load_folder)):
            os.makedirs(os.path.join(load_folder))


        if self.neural_network is None:
            X2d = np.concatenate((X2d_train, X2d_test), axis=0)
            Xnd = np.concatenate((Xnd_train, Xnd_test), axis=0)
            self.neural_network = self.fit(X2d, Xnd,
                                           epochs = nn_train_epochs, 
                                           batch_size = nn_train_batch_size,
                                           load_folder=load_folder)

        self.resolution = resolution

        self.console.log("Decoding the 2D space... 2D -> nD")

        img, img_confidence = self.get_dbm(fast_decoding_strategy, resolution, load_folder)

        self.X2d = np.concatenate((X2d_train, X2d_test), axis=0)
        self.Xnd = np.concatenate((Xnd_train.reshape((Xnd_train.shape[0], -1)), Xnd_test.reshape((Xnd_test.shape[0], -1))), axis=0)
        self.console.log("Map the 2D embedding of the data to the 2D image")

        # transform the encoded data to be in the range [0, resolution)
        X2d_train *= (resolution - 1) # type: ignore
        X2d_test *= (resolution - 1)  # type: ignore
        X2d_train = X2d_train.astype(int)
        X2d_test = X2d_test.astype(int)

        encoded_2d_train = np.zeros((len(X2d_train), 3))
        encoded_2d_test = np.zeros((len(X2d_test), 3))

        for k in range(len(X2d_train)):
            [i, j] = X2d_train[k]
            encoded_2d_train[k] = [i, j, img[i, j]]
        for k in range(len(X2d_test)):
            [i, j] = X2d_test[k]
            encoded_2d_test[k] = [i, j, img[i, j]]

        for [i, j] in X2d_test:
            img[i, j] = TEST_DATA_POINT_MARKER
            img_confidence[i, j] = 1
        for [i, j] in X2d_train:
            img[i, j] = TRAIN_DATA_POINT_MARKER
            img_confidence[i, j] = 1

        return (img, img_confidence, encoded_2d_train, encoded_2d_test)

    def _predict2dspace_(self, X2d: np.ndarray):
        """ 
        Predicts the labels for the given 2D data set.

        Args:
            X2d (np.ndarray): The 2D data set

        Returns:
            predicted_labels (np.ndarray): The predicted labels for the given 2D data set
            predicted_confidence (np.ndarray): The predicted probabilities for the given 2D data set
        """
        spaceNd = self.neural_network.decode(X2d, verbose=0)
        predictions = self.classifier.predict(spaceNd, verbose=0)                     # type: ignore
        predicted_labels = np.array([np.argmax(p) for p in predictions])
        predicted_confidence = np.array([np.max(p) for p in predictions])
        return predicted_labels, predicted_confidence, predictions

    def __transform_2d__(self, X_train: np.ndarray, X_test: np.ndarray, folder: str = DEFAULT_MODEL_PATH, projection: str = 't-SNE'):
        """ 
        Transforms the given data to 2D using a projection method.

        Args:
            X_train (np.ndarray): The training data.
            X_test (np.ndarray): The test data.
            folder (str, optional): The folder where the 2D data will be stored. Defaults to DEFAULT_MODEL_PATH.
            projection (str, optional): The projection method to be used. Defaults to 't-SNE'.

        Returns:
            X2d_train (np.ndarray): The transformed train data in 2D.
            X2d_test (np.ndarray): The transformed test data in 2D.
        """
        self.console.log(f"Transforming the data to 2D using {projection}")
        X = np.concatenate((X_train, X_test), axis=0)
        X2d = PROJECTION_METHODS[projection](X)

        self.console.log(f"Finished transforming the data to 2D using {projection}")
        X2d_train = X2d[:len(X_train)]
        X2d_test = X2d[len(X_train):]

        # rescale to [0,1]
        X2d_train, X2d_test = self.__normalize_2d__(X2d_train, X2d_test)
        # ---------------------
        if not os.path.exists(os.path.join(folder)):
            os.makedirs(os.path.join(folder))

        file_path = os.path.join(folder, TRAIN_2D_FILE_NAME)
        self.console.log("Saving the 2D data to the disk: " + file_path)
        with open(file_path, 'wb') as f:
            np.save(f, X2d_train)

        file_path = os.path.join(folder, TEST_2D_FILE_NAME)
        self.console.log("Saving the 2D data to the disk: " + file_path)
        with open(file_path, 'wb') as f:
            np.save(f, X2d_test)

        return X2d_train, X2d_test

    def __normalize_2d__(self, X2d_train: np.ndarray, X2d_test: np.ndarray):
        """ 
        Normalizes the given 2D data to [0,1].

        Args:
            X2d_train (np.ndarray): training data
            X2d_test (np.ndarray): test data

        Returns:
            X2d_train (np.ndarray): normalized training data
            X2d_test (np.ndarray): normalized test data
        """
        x_min = min(np.min(X2d_train[:, 0]), np.min(X2d_test[:, 0]))
        y_min = min(np.min(X2d_train[:, 1]), np.min(X2d_test[:, 1]))
        x_max = max(np.max(X2d_train[:, 0]), np.max(X2d_test[:, 0]))
        y_max = max(np.max(X2d_train[:, 1]), np.max(X2d_test[:, 1]))
        mins = np.array([x_min, y_min])
        ranges = np.array([x_max - x_min, y_max - y_min])

        X2d_train = (X2d_train - mins) / ranges  # type: ignore
        X2d_test = (X2d_test - mins) / ranges   # type: ignore
        return X2d_train, X2d_test



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import ceil, floor
import numpy as np
from numba import jit, njit, prange

@njit(parallel=True)
def get_nd_indices_parallel(X_nd, metric):
    """ Generates the indices of the nearest neighbors for each point in the nD space.
        Args:
            X_nd (np.ndarray): the nD dataset
            metric (function): the metric to use for calculating the distances
    """
    n_samples = X_nd.shape[0]
    dist_vector = np.zeros(n_samples, dtype=np.float64)
    indices = np.zeros((n_samples, n_samples - 1), dtype=np.int64)

    for i in range(n_samples):
        for j in prange(n_samples):
            dist_vector[j] = metric(X_nd[i], X_nd[j])

        indices[i] = np.argsort(dist_vector)[1:]  # exclude the point itself

    return indices

@njit()
def generate_point_indices(X, point, metric):
    """ Generates the indices of the nearest neighbors for a point.
        Args:
            X (np.ndarray): the (Nd or 2d) dataset
            point (np.ndarray): the (Nd or 2d) point
            metric (function): the metric to use for calculating the distances
    """
    data_samples = X.shape[0]
    dist_vector = np.zeros(data_samples, dtype=np.float64)

    for j in range(data_samples):
        dist_vector[j] = metric(point, X[j])

    return np.argsort(dist_vector)

@njit(fastmath=True)
def euclidean(x, y):
    r"""Standard euclidean distance.

    ..math::
        D(x, y) = \sqrt{\sum_i (x_i - y_i)^2}
    """
    result = 0.0
    for i in range(x.shape[0]):
        result += (x[i] - y[i]) ** 2
    return np.sqrt(result)

@jit
def get_inv_proj_error(dx, dy):
    return np.sqrt(np.linalg.norm(dx)**2 + np.linalg.norm(dy)**2)

@njit(parallel=True)
def get_proj_error_parallel(indices_source: np.ndarray, indices_embedding: np.ndarray, k: int = 10):
    """ Calculates the projection error for a given data point.
        Args:
            indices_source (np.ndarray): the indices of the point neighbors in the source (i.e. nD) space
            indices_embedding (np.ndarray): the indices of the point neighbors in the embedding (i.e. 2D) space
            k (int): the number of neighbors to consider
    """
    assert len(indices_source) == len(indices_embedding)
    n = len(indices_source)

    continuity = 0.0
    trustworthiness = 0.0

    # computing the continuity and trustworthiness errors for each point in parallel
    for i in prange(k):
        rank_2d = 0
        while indices_source[i] != indices_embedding[rank_2d]:
            rank_2d += 1

        rank_nd = 0
        while indices_source[rank_nd] != indices_embedding[i]:
            rank_nd += 1

        if rank_2d > k:
            continuity += rank_2d - k

        if rank_nd > k:
            trustworthiness += rank_nd - k

    continuity = 2 * continuity / (k * (2*n - 3*k - 1))
    trustworthiness = 2 * trustworthiness / (k * (2*n - 3*k - 1))

    return (continuity + trustworthiness) / 2

@njit()
def get_proj_error(indices_source: np.ndarray, indices_embedding: np.ndarray, k: int = 10):
    """ Calculates the projection error for a given data point.
        Args:
            indices_source (np.ndarray): the indices of the point neighbors in the source (i.e. nD) space
            indices_embedding (np.ndarray): the indices of the point neighbors in the embedding (i.e. 2D) space
            k (int): the number of neighbors to consider
    """
    assert len(indices_source) == len(indices_embedding)
    n = len(indices_source)

    continuity = 0.0
    trustworthiness = 0.0

    # computing the continuity and trustworthiness errors for each point in parallel
    for i in range(k):
        rank_2d = 0
        while indices_source[i] != indices_embedding[rank_2d]:
            rank_2d += 1

        rank_nd = 0
        while indices_source[rank_nd] != indices_embedding[i]:
            rank_nd += 1

        if rank_2d > k:
            continuity += rank_2d - k

        if rank_nd > k:
            trustworthiness += rank_nd - k

    continuity = 2 * continuity / (k * (2*n - 3*k - 1))
    trustworthiness = 2 * trustworthiness / (k * (2*n - 3*k - 1))

    return (continuity + trustworthiness) / 2

@jit
def get_pixel_priority(img, i, j, window_width, window_height, label):
    """
       Calculates the priority of decoding a chunk of pixels.
       The chunk is defined by the window size and the pixel coordinates (i,j).

    Args:
        img (np.ndarray): the image w.r.t. which the priority is calculated
        i (float): the row of the current pixel
        j (float): the column of the current pixel
        window_width (float), window_height (float): the window size (i.e. the chunk size)
        label (any): the label of the current pixel

    Returns:
        priority (float): the priority of decoding the chunk in range [0,1] or -1 if the chunk does not need to be decoded
    """
    resolution = img.shape[0]
    # getting the 4 neighbors
    w = (window_width - 1) / 2
    h = (window_height - 1)/ 2
    neighbors = []

    if i - h - 1 >= 0:
        neighbors.append(img[int(i - h - 1), int(j)])
    if i + h + 1 < resolution:
        neighbors.append(img[int(i + h + 1), int(j)])

    if j - w - 1 >= 0:
        neighbors.append(img[int(i), int(j - w - 1)])
    if j + w + 1 < resolution:
        neighbors.append(img[int(i), int(j + w + 1)])

    cost = 0
    for neighbor in neighbors:
        if neighbor != label:
            cost += 1


    cost /= len(neighbors)
    cost *= window_width * window_height
    
    if cost == 0:
        return -1

    return 1/cost

@njit
def binary_split(i, j, W, H):
    Wc, Wf = ceil(W/2), floor(W/2)
    Hc, Hf = ceil(H/2), floor(H/2)
    slack_w = 0 if Wc == Wf else 0.5
    slack_h = 0 if Hc == Hf else 0.5
    c_i = i + slack_h
    c_j = j + slack_w
    representatives = [(c_j - Wc / 2, c_i - Hc / 2), 
                       (c_j - Wc / 2, c_i + Hf / 2), 
                       (c_j + Wf / 2, c_i - Hc / 2),
                       (c_j + Wf / 2, c_i + Hf / 2)]
    sizes = [(Wc, Hc), (Wc, Hf), (Wf, Hc), (Wf, Hf)]
    return representatives, sizes

@njit(parallel=True)
def get_projection_errors_using_inverse_projection(Xnd: np.ndarray, X2d: np.ndarray, spaceNd: np.ndarray, space2d: np.ndarray, progress, k: int = 10):

    n_points = len(space2d)
    data_samples = len(X2d)
    errors = np.zeros(n_points, dtype=np.float64)

    for index in prange(n_points):
        indices_embedded, indices_source = np.zeros(data_samples, dtype=np.int64), np.zeros(data_samples, dtype=np.int64)
        indices_embedded = generate_point_indices(X2d, space2d[index], metric=euclidean)
        indices_source = generate_point_indices(Xnd, spaceNd[index], metric=euclidean)
        errors[index] = get_proj_error(indices_source, indices_embedded, k=k)
        progress.update(1)

    return errors

@njit
def generate_windows(window_size: int, initial_resolution: int, resolution: int = 1024):

    indexes = [((i * window_size + window_size / 2 - 0.5), (j * window_size + window_size / 2 - 0.5)) for i in range(initial_resolution) for j in range(initial_resolution)]
    
    border_indexes = [(-1,-1), (-1, resolution), (resolution, -1), (resolution, resolution)]
    for i in range(initial_resolution):
        p = (i * window_size + window_size / 2 - 0.5)
        border_indexes += [(-1, p), (p, -1), (p, resolution), (resolution, p)]
    
    sizes = [(window_size, window_size)] * len(indexes)
    if initial_resolution * window_size == resolution:
        return indexes, sizes, border_indexes
    
    slack = (resolution - initial_resolution * window_size) / 2
    p = resolution - slack - 0.5
       
    indexes += [(i * window_size + window_size / 2 - 0.5, p) for i in range(initial_resolution)]
    sizes += [(window_size, (resolution - initial_resolution * window_size))] * initial_resolution 
    indexes += [(p, i * window_size + window_size / 2 - 0.5) for i in range(initial_resolution)]
    sizes += [((resolution - initial_resolution * window_size), window_size)] * initial_resolution
    
    indexes += [(p, p)]
    sizes += [((resolution - initial_resolution * window_size), (resolution - initial_resolution * window_size))]
    
    border_indexes += [(-1, p), (p, -1), (p, resolution), (resolution, p)]
    
    return indexes, sizes, border_indexes

@njit
def get_window_borders(x, y, w, h):
    # returns the borders of the window by its center and size
    # x, y - center of the window
    # w, h - size of the window
    # returns left, right, top, bottom
    return int(x - (w - 1) / 2), int(x + (w - 1) / 2), int(y - (h - 1) / 2), int(y + (h - 1) / 2)

def get_tasks_with_same_priority(priority_queue):
    # take the highest priority task
    priority, item = priority_queue.get()
    # getting all the items with the same priority
    items = [item]
    if priority_queue.empty():
        return items
        
    next_priority, next_item = priority_queue.get()
    while priority == next_priority:
        items.append(next_item)
        if priority_queue.empty():
            break
        next_priority, next_item = priority_queue.get()
        
    # putting back the last item in the queue
    if priority != next_priority:
        priority_queue.put((next_priority, next_item))
            
    return items

@njit 
def get_split_position(x1: float, x2:float, bound: float, c11: float, c12: float, c21: float, c22: float) -> int | None:
    assert(x1 != x2)

    a, b = (c11 - c12) / (x1 - x2), c11 - x1 * ((c11 - c12) / (x1 - x2)) 
    c, d = (c21 - c22) / (x1 - x2), c21 - x1 * ((c21 - c22) / (x1 - x2))

    if a == c:
        return None
    
    boundary = round((d - b) / (a - c))

    if (bound < x1) and (bound + 1 < boundary < x1):
        return boundary
    if (bound > x1) and (x1 < boundary < bound - 1):
        return boundary
    
    return None

def get_confidence_splits(img, conf_img, img_indexes, c_y, c_x, top, bottom, left, right):
    resolution = img.shape[0]
    i, j = int(c_y), int(c_x)
    label = img[i, j]
    c11 = conf_img[i, j][label]

    splits_x, splits_y = [], []    
    
    vertical = [top] if top >= 0 else []
    vertical += [bottom] if bottom < resolution else []
    horizontal = [left] if left >= 0 else []
    horizontal += [right] if right < resolution else []
    
    for k in vertical:
        new_label = img[k, j]
        (x, y) = img_indexes[k, j]
        if new_label != label:
            split = get_split_position(c_y, y, k, c11, conf_img[y, j][label], conf_img[i, j][new_label], conf_img[y, j][new_label])
            if split is not None:
                splits_y.append(split)

    for k in horizontal:
        new_label = img[i, k]
        (x, y) = img_indexes[i, k]
        if new_label != label:
            split = get_split_position(c_x, x, k, c11, conf_img[i, x][label], conf_img[i, j][new_label], conf_img[i, x][new_label])
            if split is not None:
                splits_x.append(split)
                
    return splits_x, splits_y

def get_confidence_based_split(img, conf_img, img_indexes, i, j, W, H):
    c_y, c_x = i, j
    left, right, top, bottom = get_window_borders(c_x, c_y, W, H)
    # going to the nearest border outside of the window
    left -= 1
    right += 1
    top -= 1
    bottom += 1
    splits_x, splits_y = get_confidence_splits(img, conf_img, img_indexes, c_y, c_x, top, bottom, left, right)
    
    if len(splits_x) == 0 and len(splits_y) == 0:
        return binary_split(c_y, c_x, W, H)
    
    return confidence_based_split(splits_x, splits_y, c_y, c_x, W, H, top, bottom, left, right)

def confidence_based_split(splits_x, splits_y, c_y, c_x, W, H, top, bottom, left, right):
    markers_x, markers_y = [], []
    
    if len(splits_x) == 0:
        Wc, Wf = ceil(W/2), floor(W/2)
        slack_w = 0 if Wc == Wf else 0.5
        c_j = c_x + slack_w
        markers_x = [(c_j - Wc / 2, Wc), (c_j + Wf / 2, Wf)]    
        
    if len(splits_y) == 0:
        Hc, Hf = ceil(H/2), floor(H/2)
        slack_h = 0 if Hc == Hf else 0.5
        c_i = c_y + slack_h
        markers_y = [(c_i - Hc / 2, Hc), (c_i + Hf / 2, Hf)]
    
    if len(splits_x) == 1:
        x = splits_x[0]
        markers_x = [((x + left + 1) / 2, abs(x - left)), ((x + right) / 2, abs(x - right) - 1)]
    if len(splits_y) == 1:
        y = splits_y[0]
        markers_y = [((y + top + 1) / 2, abs(y - top)), ((y + bottom) / 2, abs(y - bottom) - 1)]
        
    if len(splits_x) == 2:
        x1, x2 = splits_x
        markers_x = [((x1 + left + 1) / 2, abs(x1 - left)), ((x1 + x2) / 2, abs(x1 - x2) - 1), ((x2 + right - 1) / 2, abs(x2 - right))]
    if len(splits_y) == 2:
        y1, y2 = splits_y
        markers_y = [((y1 + top + 1) / 2, abs(y1 - top)), ((y1 + y2) / 2, abs(y1 - y2) - 1), ((y2 + bottom - 1) / 2, abs(y2 - bottom))]
        
    representatives, sizes = [], []
    for (y, h) in markers_y:
        for (x, w) in markers_x:
            representatives.append((x, y))
            sizes.append((w, h))
    
    return representatives, sizes



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import sqrt
import tensorflow as tf
import os
import numpy as np
import json
import matplotlib.pyplot as plt

from ..Logger import Logger, LoggerInterface

TRAINING_HISTORY_FILE_NAME = "history.json"
SEED = 42

class AbstractNN:
    """ 
    Defines an abstract class for the neural networks used by the Decision Boundary Mapper.

    Public methods:
        load \n
        save \n
        show_predictions \n

    Methods to be implemented by the classes that inherit from this class.
        encode: Encodes the input data.
    """

    def __init__(self,
                 folder_path: str,
                 nn_name: str,
                 logger: LoggerInterface | None = None):
        """ 
        Initializes an NN object.

        Args:
            folder_path (str): the folder path where the model will be saved/loaded.
            nn_name (str): Name of the neural network that uses the classifier.
            logger (LoggerInterface, optional): Defaults to console logging.

        Raises:
            Exception: If the classifier is not provided.
        """
        if logger is not None:
            self.console = logger
        else:
            self.console = Logger(name=nn_name)

        self.save_folder_path = folder_path
        self.nn_name = nn_name
        self.neural_network = None

        try:
            self.load()
        except Exception as e:
            self.console.warn("Model not found, {}".format(e))
            self.console.warn("The model will be built and trained from scratch.")

    def load(self):
        """
            Loads an auto encoder from the specified folder path. With the .h5 extension.
            Args:
                folder_path (str): The path to the folder where the model is saved.
        """
        try:
            self.neural_network = tf.keras.models.load_model(os.path.join(self.save_folder_path, self.nn_name), compile=False)
            self.console.log("NN loaded successfully")
        except Exception as e:
            self.console.log(
                f"NN not found. Please check the path folder {self.save_folder_path} and make sure the model is saved there")
            raise e

    def save(self, history=None):
        """
            Saves the model to the specified folder path. With the .tf extension.
        """
        folder_path = self.save_folder_path
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        path = os.path.join(folder_path, self.neural_network.name)  # type: ignore
        self.neural_network.save(path, save_format="tf")            # type: ignore
        self.console.log(f"Model saved to {folder_path}")

        if history is None:
            return

        with open(os.path.join(folder_path, TRAINING_HISTORY_FILE_NAME), "w") as f:
            f.write(json.dumps(history.history))
            self.console.log(f"History saved to {folder_path}")

    def show_predictions(self, dataNd: np.ndarray, labels: np.ndarray, data2d: np.ndarray | None = None, n_samples: int = 16):
        """ Shows the predictions of the model. By taking first n_samples data points and comparing them to the actual labels.

        Args:
            dataNd (np.ndarray): N-dimensional data points.
            data2d (np.ndarray, optional): 2-dimensional data points. Defaults to None.
            labels (np.ndarray): Actual labels of the data points
            n_samples (int, optional): Number of samples to show. Defaults to 16.
        """
        if data2d is None:
            data2d = self.encode(dataNd)

        decoded, predicted = self.decode(data2d)
        predicted_labels = np.argmax(predicted, axis=1)

        plt.figure(figsize=(20, 10))
        m = int(sqrt(n_samples))
        n = m * m

        for i in range(1, 2*n, 2):
            plt.subplot(m, 2*m, i)
            plt.imshow(dataNd[i], cmap='gray')
            plt.title(f"Actual: {labels[i]}", color='green' if predicted_labels[i]== labels[i] else 'red', fontsize=12)
            plt.axis('off')
            plt.subplot(m, 2*m, i+1)
            plt.imshow(decoded[i], cmap='gray')
            plt.title(f"Predicted: {predicted_labels[i]}", color='green' if predicted_labels[i] == labels[i] else 'red', fontsize=12)
            plt.axis('off')

        plt.show()

    def decode(self, data: np.ndarray, verbose: int = 0) -> np.ndarray:
        """ Decodes the data points.

        Args:
            data (np.ndarray): The data points to decode.
            verbose (int, optional): Verbosity mode. Defaults to 0.

        Returns:
            Xnd, labels_predictions: The decoded data points and the predictions of the classifier.
        """
        return self.neural_network.predict(data, verbose=verbose)  # type: ignore

    def encode(self, data: np.ndarray, verbose: int = 0) -> np.ndarray:
        """ Encodes the data points.

        Args:
            data (np.ndarray): The data points to encode.
            verbose (int, optional): Verbosity mode. Defaults to 0.

        Returns:
            np.ndarray: The encoded 2D data points.
        """
        raise NotImplementedError("The encode method is not implemented.")



# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#GLOBAL CONSTANTS are defined here
TRAIN_DATA_POINT_MARKER = -1
TEST_DATA_POINT_MARKER = -2

TRAIN_2D_FILE_NAME = "train_2d.npy"
TEST_2D_FILE_NAME = "test_2d.npy"

INVERSE_PROJECTION_ERRORS_FILE = "inverse_projection_errors.npy"
PROJECTION_ERRORS_INTERPOLATED_FILE = "projection_errors_interpolated.npy"
PROJECTION_ERRORS_INVERSE_PROJECTION_FILE = "projection_errors_inv_projection.npy"


#GUI CONSTANTS
APP_FONT = "Helvetica 12"
APP_FONT_BOLD = "Helvetica 12 bold"
BLACK_COLOR = "#252526"
APP_PRIMARY_COLOR = "#007acc"
WHITE_COLOR = "#ffffff"
RED_COLOR = "#ff0000"
GREEN_COLOR = "#00ff00"
YELLOW_COLOR = "#ffff00"
RIGHTS_MESSAGE_1 = "© 2023 Cristian Grosu. All rights reserved."
RIGHTS_MESSAGE_2 = "Made by Cristian Grosu for Utrecht University Master Thesis in 2023"


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy as np
import json
import threading
from math import sqrt
from datetime import datetime
import shutil
from matplotlib.patches import Patch, Circle
from matplotlib.offsetbox import OffsetImage, AnnotationBbox, TextArea

from .. import Logger
from ..utils import TRAIN_DATA_POINT_MARKER, TEST_DATA_POINT_MARKER, TRAIN_2D_FILE_NAME, TEST_2D_FILE_NAME, INVERSE_PROJECTION_ERRORS_FILE, PROJECTION_ERRORS_INTERPOLATED_FILE, PROJECTION_ERRORS_INVERSE_PROJECTION_FILE, get_latest_created_file_from_folder, run_timer

CLASSIFIER_PERFORMANCE_HISTORY_FILE = "classifier_performance.log"
CLASSIFIER_REFIT_FOLDER = "refit_classifier"
CLASSIFIER_STACKED_FOLDER = "stacked_classifier"
CLASSIFIER_STACKED_LABELS_FILE = "classifier_old_labels.npy"
CLASSIFIER_STACKED_LABELS_CHANGES_FILE = "classifier_old_labels_changes.npy"
CLASSIFIER_STACKED_BOUNDARY_MAP_FILE = "classifier_old_boundary_map.npy"
CLASSIFIER_STACKED_CONFIDENCE_MAP_FILE = "classifier_old_boundary_map_confidence.npy"

PLOT_SNAPSHOTS_FOLDER = "plot_snapshots"

LABELS_CHANGES_FILE = "label_changes.json"

USER_ALLOWED_INTERACTION_ITERATIONS = 5

TIMER_SOUND_FILE_PATH = os.path.join(os.path.dirname(__file__), "assets", "warning-sound.mp3")
TIMER_DURATION = int(3 * 60) # seconds
TIMER_WARNING_INTERVAL = 30 # seconds

EPOCHS_FOR_REFIT = 20
EPOCHS_FOR_REFIT_RANGE = (1, 100)

class DBMPlotterController:
    def __init__(self,
                logger,
                dbm_model,
                img, img_confidence,
                X_train, Y_train,
                X_test, Y_test,
                X_train_2d, X_test_2d,
                encoded_train, encoded_test,
                save_folder,
                projection_technique,
                gui
                ):
        
        if logger is None:
            self.console = Logger(name="DBMPlotterController")
        else:
            self.console = logger
            
        # folder where to save the changes made to the data by the user
        self.save_folder = save_folder
        # projection technique used to generate the DBM
        self.projection_technique = projection_technique
        if self.projection_technique is not None:
            self.save_folder = os.path.join(self.save_folder, self.projection_technique)
        
        self.inverse_projection_errors = None
        self.projection_errors = None
        self.positions_of_labels_changes = ([], [], [])

        self.dbm_model = dbm_model
        self.img = img
        self.img_confidence = img_confidence
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_test = X_test
        self.Y_test = Y_test
        self.X_train_2d = X_train_2d
        self.X_test_2d = X_test_2d
        self.encoded_train = encoded_train
        self.encoded_test = encoded_test
        self.gui = gui # reference to the gui that uses the controller
        self.user_allowed_interaction_iterations = USER_ALLOWED_INTERACTION_ITERATIONS
        
        # -------------- Create folder structure --------------------
        folders_to_create = [
            CLASSIFIER_REFIT_FOLDER,
            CLASSIFIER_STACKED_FOLDER,
            PLOT_SNAPSHOTS_FOLDER
        ]
        for folder in folders_to_create:
            dir = os.path.join(self.save_folder, folder)
            if not os.path.exists(dir):
                os.makedirs(dir)
        
        # ----------------------------------------------------------------
        self.stop_timer_event = threading.Event()
        
        self.initialize()
        
    def initialize(self):
        self.train_mapper, self.test_mapper = self.generate_encoded_mapping()
        # --------------------- Others ------------------------------
        self.expert_updates_labels_mapper = {}
        self.motion_event_cid = None
        self.click_event_cid = None
        self.key_event_cid = None
        self.release_event_cid = None
        self.current_selected_point = None
        self.current_selected_point_assigned_label = None
        self.update_labels_by_circle_select = True
        self.update_labels_circle = None
    
    def clear_resources(self):
        # stop the timer if there is any
        self.stop_timer()
        
        files_to_delete = [
            CLASSIFIER_PERFORMANCE_HISTORY_FILE,
            LABELS_CHANGES_FILE,
            CLASSIFIER_STACKED_BOUNDARY_MAP_FILE,
            CLASSIFIER_STACKED_CONFIDENCE_MAP_FILE,
            CLASSIFIER_STACKED_LABELS_FILE,
            CLASSIFIER_STACKED_LABELS_CHANGES_FILE
        ]

        for file in files_to_delete:
            file = os.path.join(self.save_folder, file)
            if os.path.exists(file):
                os.remove(file)

        folders_to_delete = [
            CLASSIFIER_STACKED_FOLDER
        ]

        for folder in folders_to_delete:
            folder = os.path.join(self.save_folder, folder)
            if os.path.exists(folder):
                shutil.rmtree(folder)

    def build_2D_image(self, colors_mapper, class_name_mapper=lambda x: str(x)):
        """Combines the img and the img_confidence into a single image.

        Args:
            img (np.ndarray): Label image.
            img_confidence (np.ndarray): Confidence image.
            colors_mapper (dict): Mapper of labels to colors.
            class_name_mapper (function, optional): Describes how to map the data labels. Defaults to lambda x:str(x).
            
        Returns:
            np.ndarray: The combined image.
            patches: The legend patches.
        """
        img = self.img
        color_img = np.zeros((img.shape[0], img.shape[1], 3))

        for i, j in np.ndindex(img.shape):
            color_img[i][j] = colors_mapper[img[i][j]]
        values = np.unique(img)

        patches = []
        for value in values:
            color = colors_mapper[value]
            if value == TRAIN_DATA_POINT_MARKER:
                label = "Original train data"
            elif value == TEST_DATA_POINT_MARKER:
                label = "Original test data"
            else:
                label = f"Value region: {class_name_mapper(value)}"

            patches.append(Patch(color=color, label=label))

        self.color_img = color_img
        self.legend = patches

        return color_img, patches

    def generate_encoded_mapping(self):
        """Generates a mapping of the encoded data to the original data.

        Returns:
            train_mapper (dict): Mapping of the encoded train data to the original train data.
            test_mapper (dict): Mapping of the encoded test data to the original test data.
        """
        train_mapper = {}
        for k in range(len(self.encoded_train)):
            [i, j, _] = self.encoded_train[k]
            train_mapper[f"{int(i)} {int(j)}"] = k

        test_mapper = {}
        for k in range(len(self.encoded_test)):
            [i, j, _] = self.encoded_test[k]
            test_mapper[f"{int(i)} {int(j)}"] = k

        return train_mapper, test_mapper
    
    def get_classifier_performance_history(self):
        path = os.path.join(self.save_folder, CLASSIFIER_PERFORMANCE_HISTORY_FILE)
        if not os.path.isfile(path):
            raise Exception("Classifier performance history file not found.")

        times, accuracies, losses = [], [], []
        with open(path, "r") as f:
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0:
                    continue
                _, time, _, acc, _, loss = line.replace("\n", "").replace("%", "").split(" ")
                times.append(time)
                accuracies.append(float(acc))
                losses.append(float(loss))
        return times, accuracies, losses
    
    def compute_classifier_metrics(self, epochs):
        self.console.log("Evaluating classifier...")
        loss, accuracy = self.dbm_model.classifier.evaluate(self.X_test, self.Y_test, verbose=0)
        self.console.log(f"Classifier Accuracy: {(100 * accuracy):.2f}%  Loss: {loss:.2f}")

        path = os.path.join(self.save_folder, CLASSIFIER_PERFORMANCE_HISTORY_FILE)

        with open(path, "a") as f:
            time = datetime.now().strftime("%D %H:%M:%S")
            message = f"Accuracy: {(100 * accuracy):.2f}% Loss: {loss:.2f}"
            f.write(f"{time} {message}\n")
        return accuracy, loss
    
    def pop_classifier_evaluation(self):
        path = os.path.join(self.save_folder, CLASSIFIER_PERFORMANCE_HISTORY_FILE)
        # removing the last line
        with open(path, "r") as f:
            lines = f.readlines()
            lines = lines[:-1]
        with open(path, "w") as f:
            f.write("".join(lines))

        if len(lines) == 0:
            return None, None

        last_line = lines[-1].replace("\n", "")
        accuracy, loss = float(last_line.split("Accuracy: ")[1].split("%")[0]), float(last_line.split("Loss: ")[1])
        return accuracy, loss
    
    def load_2d_projection(self):
        if os.path.exists(os.path.join(self.save_folder, TRAIN_2D_FILE_NAME)) and os.path.exists(os.path.join(self.save_folder, TEST_2D_FILE_NAME)):
            X2d_train = np.load(os.path.join(self.save_folder, TRAIN_2D_FILE_NAME))
            X2d_test = np.load(os.path.join(self.save_folder, TEST_2D_FILE_NAME))
            return X2d_train, X2d_test
        return None, None
    
    def transform_changes(self, Y_train, expert_updates_labels_mapper, positions_of_labels_changes):
        """
        Returns:
            position_changes (dict): a dictionary with old position as key and the new position as value
            label_changes (dict): a dictionary with old and new positions as key and the new label as value
        """
        Y = np.copy(Y_train)
        labels_changes = {}

        pos_x, pos_y, alphas = positions_of_labels_changes

        ALPHA_DECAY_ON_UPDATE = 0.05
        alphas = [alpha - ALPHA_DECAY_ON_UPDATE if alpha > ALPHA_DECAY_ON_UPDATE else ALPHA_DECAY_ON_UPDATE for alpha in alphas]

        for pos in expert_updates_labels_mapper:
            k = self.train_mapper[pos]
            pos_x = np.append(pos_x, int(pos.split(" ")[1]))
            pos_y = np.append(pos_y, int(pos.split(" ")[0]))
            alphas = np.append(alphas, 1)
            y = expert_updates_labels_mapper[pos][0]
            Y[k] = y
            labels_changes[pos] = expert_updates_labels_mapper[pos][0]

        positions_of_labels_changes = (pos_x, pos_y, alphas)

        return Y, labels_changes, positions_of_labels_changes
    
    def save_labels_changes(self, folder: str = "tmp", label_changes={}):
        if not os.path.exists(folder):
            os.path.makedirs(folder)

        with open(os.path.join(folder, LABELS_CHANGES_FILE), "a") as f:
            f.write("\n" + "-" * 50 + "\n")
            json.dump(label_changes, f, indent=2)

    def compute_inverse_projection_errors(self):
        possible_path = os.path.join(self.save_folder, INVERSE_PROJECTION_ERRORS_FILE)
        # try to get projection errors from cache first
        if os.path.exists(possible_path):
            self.inverse_projection_errors = np.load(possible_path)
        else:
            self.inverse_projection_errors = self.dbm_model.generate_inverse_projection_errors(save_folder=self.save_folder, resolution=len(self.img))
            
    def compute_projection_errors(self, type = "non_interpolated"):
        if type == "interpolated":
            use_interpolation = True
            file = PROJECTION_ERRORS_INTERPOLATED_FILE
        elif type == "non_interpolated":
            use_interpolation = False
            file = PROJECTION_ERRORS_INVERSE_PROJECTION_FILE
            
        possible_path = os.path.join(self.save_folder, file)

        # try to get projection errors from cache first
        if os.path.exists(possible_path):
            self.projection_errors = np.load(possible_path)
        else:
            self.projection_errors = self.dbm_model.generate_projection_errors(use_interpolation=use_interpolation, save_folder=self.save_folder)

    def undo_changes(self):
        if not os.path.exists(os.path.join(self.save_folder, CLASSIFIER_STACKED_FOLDER)):
            raise Exception("Can not undo changes, no previous model found")

        if not os.path.exists(os.path.join(self.save_folder, CLASSIFIER_STACKED_LABELS_FILE)):
            raise Exception("Can not undo changes, no previous labels found")

        if not os.path.exists(os.path.join(self.save_folder, CLASSIFIER_STACKED_BOUNDARY_MAP_FILE)):
            raise Exception("Can not undo changes, no previous boundary map found")

        if not os.path.exists(os.path.join(self.save_folder, CLASSIFIER_STACKED_CONFIDENCE_MAP_FILE)):
            raise Exception("Can not undo changes, no previous confidence map found")
        
        if not os.path.exists(os.path.join(self.save_folder, CLASSIFIER_STACKED_LABELS_CHANGES_FILE)):
            raise Exception("Can not undo changes, no previous labels changes found")
        
        if not os.path.exists(os.path.join(self.save_folder, CLASSIFIER_STACKED_FOLDER)):
            raise Exception("Can not undo changes, no previous model found")

        self.dbm_model.load_classifier(os.path.join(self.save_folder, CLASSIFIER_STACKED_FOLDER))

        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_LABELS_FILE), "rb") as f:
            self.Y_train = np.load(f)
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_BOUNDARY_MAP_FILE), "rb") as f:
            self.img = np.load(f)
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_CONFIDENCE_MAP_FILE), "rb") as f:
            self.img_confidence = np.load(f)
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_LABELS_CHANGES_FILE), "rb") as f:
            self.positions_of_labels_changes = np.load(f)
            
        files_to_delete = [
            CLASSIFIER_STACKED_LABELS_FILE,
            CLASSIFIER_STACKED_BOUNDARY_MAP_FILE,
            CLASSIFIER_STACKED_CONFIDENCE_MAP_FILE,
            CLASSIFIER_STACKED_LABELS_CHANGES_FILE
        ]

        for file in files_to_delete:
            file = os.path.join(self.save_folder, file)
            if os.path.exists(file):
                os.remove(file)
                
        # revoke the latest plot snapshot
        if not os.path.exists(os.path.join(self.save_folder, PLOT_SNAPSHOTS_FOLDER)):
            return
        
        plot_snapshots_folder = os.path.join(self.save_folder, PLOT_SNAPSHOTS_FOLDER)
        try:
            latest_plot_snapshot_file_path = get_latest_created_file_from_folder(plot_snapshots_folder)
            os.rename(
                      latest_plot_snapshot_file_path,
                      latest_plot_snapshot_file_path.replace(".png", "_revoked.png")
                      )
        except FileExistsError:
            self.console.error("Couldn't revoke the latest plot snapshot since it is already revoked.")
        except Exception as e:
            self.console.error(f"Error while revoking the latest plot snapshot: {str(e)}")

    def mix_image(self, show_color_map, show_confidence, show_inverse_projection_errors, show_projection_errors):
        color_img = np.zeros((self.img.shape[0], self.img.shape[1], 3))
        alphas = 1 + np.zeros((self.img.shape[0], self.img.shape[1]))
        img = np.zeros((self.img.shape[0], self.img.shape[1], 4))

        if show_color_map:
            color_img = self.color_img

        if show_confidence:
            alphas = alphas * self.img_confidence
        if show_inverse_projection_errors:
            alphas = alphas * (1 - self.inverse_projection_errors)

        if show_projection_errors and self.projection_errors is not None:
            alphas = alphas * (1 - self.projection_errors)

        img[:, :, :3] = color_img
        img[:, :, 3] = alphas
        return img
    
    def get_encoded_train_data(self):
        return self.encoded_train
    
    def get_positions_of_labels_changes(self):
        return self.positions_of_labels_changes
    
    def regenerate_boundary_map(self, Y_transformed, fast_decoding_strategy):
        if self.projection_technique is None:
            dbm_info = self.dbm_model.generate_boundary_map(
                self.X_train,
                Y_transformed,
                self.X_test,
                self.Y_test,
                resolution=len(self.img),
                fast_decoding_strategy=fast_decoding_strategy,
                load_folder=self.save_folder
            )
        else:
            if self.X_train_2d is None or self.X_test_2d is None:
                self.X_train_2d, self.X_test_2d = self.load_2d_projection()
            dbm_info = self.dbm_model.generate_boundary_map(
                Xnd_train=self.X_train,
                Xnd_test=self.X_test,
                X2d_train=self.X_train_2d,
                X2d_test=self.X_test_2d,
                resolution=len(self.img),
                fast_decoding_strategy=fast_decoding_strategy,
                load_folder=self.save_folder,
                projection=self.projection_technique
            )
        
        img, img_confidence, encoded_train, encoded_test = dbm_info
        
        self.img = img
        self.img_confidence = img_confidence
        self.encoded_train = encoded_train
        self.encoded_test = encoded_test
        self.Y_train = Y_transformed
        self.initialize()
        
    def apply_labels_changes(self, decoding_strategy, epochs = None):
        if self.user_allowed_interaction_iterations <= 0:
            message = "Max number of iterations reached! Applying labels changes is not allowed anymore!"
            self.console.error(message)
            self.updates_logger.error(message)
        
        num_changes = len(self.expert_updates_labels_mapper)
        
        if num_changes == 0:
            message = "No changes to apply!"
            self.console.error(message)
            self.updates_logger.error(message)
            return
        
        # store the changes done so far so we can restore them when needed
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_LABELS_CHANGES_FILE), "wb") as f:
            np.save(f, self.positions_of_labels_changes)

        Y_transformed, label_changes, positions_of_labels_changes = self.transform_changes(self.Y_train, self.expert_updates_labels_mapper, self.positions_of_labels_changes)
        
        if len(label_changes) > 0.8 * len(self.Y_train):
            message = "The amount of changes can not be more than 80% of the training set in one iteration"
            self.console.error(message)
            self.updates_logger.error(message)
            return
        
        # stop the user timer
        self.stop_timer()
        
        # block the user from applying changes
        self.gui.window["-APPLY CHANGES SECTION-"].update(visible=False)
        
        self.positions_of_labels_changes = positions_of_labels_changes

        self.console.log("Saving changes to a local folder...")
        self.save_labels_changes(self.save_folder, label_changes=label_changes)

        self.updates_logger.log("Applying changes... This might take some time...")

        save_folder = os.path.join(self.save_folder, CLASSIFIER_REFIT_FOLDER)

        # store the old model so we can restore it when needed
        self.dbm_model.save_classifier(save_folder=os.path.join(self.save_folder, CLASSIFIER_STACKED_FOLDER))
        # store the old labels so we can restore them when needed
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_LABELS_FILE), "wb") as f:
            np.save(f, self.Y_train)

        # store the old decision boundary map and confidence map so we can restore them when needed
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_BOUNDARY_MAP_FILE), "wb") as f:
            np.save(f, self.img)
        with open(os.path.join(self.save_folder, CLASSIFIER_STACKED_CONFIDENCE_MAP_FILE), "wb") as f:
            np.save(f, self.img_confidence)

        # store the plot presented when the user applies the changes
        current_time = datetime.now().strftime("%D %H:%M:%S").replace(" ", "_").replace("/", "_")
        self.gui.fig.savefig(os.path.join(self.save_folder, PLOT_SNAPSHOTS_FOLDER, f"{current_time}.png"))

        if epochs is None:
            epochs = EPOCHS_FOR_REFIT
        
        self.updates_logger.log(f"The classifier will be retrained for {epochs} epochs")

        self.dbm_model.refit_classifier(self.X_train, Y_transformed, save_folder=save_folder, epochs=epochs)
        self.regenerate_boundary_map(Y_transformed, decoding_strategy)

        self.updates_logger.log("Changes applied successfully!")

        # decrease the number of user allowed interactions
        self.user_allowed_interaction_iterations -= 1
        
        # if user is not allowed interact block the buttons and show a informative message
        if self.user_allowed_interaction_iterations <= 0:
            self.gui.window["-USAGE ITERATIONS LEFT TEXT-"].update("Max number of iterations reached!\nApplying labels changes is not allowed anymore!", text_color='red')
            self.gui.window["-APPLY CHANGES-"].hide_row()
            self.gui.window["-UNDO CHANGES-"].hide_row()
            self.timer_ui.hide_row()
            self.gui.window["-APPLY CHANGES SECTION-"].update(visible=True)
            return
        
        # update the number of user allowed interactions in the gui
        self.gui.window["-USAGE ITERATIONS LEFT TEXT-"].update(f"Usage iterations left: {self.user_allowed_interaction_iterations}")
               
        # restart the timer
        self.start_timer(self.timer_ui)
        
        self.gui.window["-APPLY CHANGES SECTION-"].update(visible=True)
        
    def start_timer(self, ui_component):
        # create a new timer that will update the ui_component
        # set the stop_timer_event in order to stop the timer when needed
        self.stop_timer_event = threading.Event()
        self.timer_ui = ui_component
        self.timer_ui.update(visible=True)
        
        def timer_update_callback(msg, is_time_up):
            text_color = "red" if is_time_up else "green"
            self.timer_ui.update(msg, text_color=text_color)
            self.gui.window.refresh()
            
        threading.Thread(target=run_timer,
                        args=(
                            timer_update_callback,
                            self.stop_timer_event,
                            TIMER_DURATION,
                            TIMER_WARNING_INTERVAL,
                            TIMER_SOUND_FILE_PATH,
                            True,
                            )).start()
       
    def stop_timer(self):
        self.stop_timer_event.set()
    
    def set_dbm_model_logger(self, logger):
        self.dbm_model.console = logger
    
    def set_updates_logger(self, logger):
        self.updates_logger = logger
    
    def build_annotation_mapper(self, fig, ax, connect_click_event=True):
        """ Builds the annotation mapper.
            This is used to display the data point label when hovering over the decision boundary.
        """
        image = OffsetImage(self.X_train[0], zoom=2, cmap="gray")
        label = TextArea("Data point label: None")

        annImage = AnnotationBbox(image, (0, 0), xybox=(50., 50.), xycoords='data', boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))
        annLabels = AnnotationBbox(label, (0, 0), xybox=(50., 50.), xycoords='data', boxcoords="offset points",  pad=0.3,  arrowprops=dict(arrowstyle="->"))

        ax.add_artist(annImage)
        ax.add_artist(annLabels)
        annImage.set_visible(False)
        annLabels.set_visible(False)

        fig_width, fig_height = fig.get_size_inches() * fig.dpi

        def display_annotation(event):
            """Displays the annotation box when hovering over the decision boundary based on the mouse position."""
            if event.inaxes == None:
                return

            j, i = int(event.xdata), int(event.ydata)

            # change the annotation box position relative to mouse.
            ws = (event.x > fig_width/2.)*-1 + (event.x <= fig_width/2.)
            hs = (event.y > fig_height/2.)*-1 + (event.y <= fig_height/2.)
            annImage.xybox = (50. * ws, 50. * hs)
            annLabels.xybox = (-50. * ws, 50. * hs)
            # make annotation box visible
            annImage.set_visible(True)
            # place it at the position of the event scatter point
            annImage.xy = (j, i)

            x_data, y_data = find_data_point(i, j)
            if x_data is not None:
                annImage.set_visible(True)
                resize_factor = int(len(x_data) / 50 + 1)
                annImage.xybox = (50. * ws * resize_factor, 50. * hs * resize_factor)
                image.set_data(x_data)
            else:
                annImage.set_visible(False)

            if y_data is not None:
                annLabels.set_visible(True)
                annLabels.xy = (j, i)
                label.set_text(f"{y_data}")
            else:
                annLabels.set_visible(False)

            fig.canvas.draw_idle()

        def find_data_point(i, j):
            # search for the data point in the encoded train data
            if self.img[i][j] == TRAIN_DATA_POINT_MARKER:
                k = self.train_mapper[f"{i} {j}"]
                if f"{i} {j}" in self.expert_updates_labels_mapper:
                    l = self.expert_updates_labels_mapper[f"{i} {j}"][0]
                    return self.X_train[k], f"Label {self.Y_train[k]} \nClassifier label: {int(self.encoded_train[k][2])} \nExpert label: {l}"
                return self.X_train[k], f"Label: {self.Y_train[k]} \nClassifier label: {int(self.encoded_train[k][2])}"

            # search for the data point in the encoded test data
            if self.img[i][j] == TEST_DATA_POINT_MARKER:
                k = self.test_mapper[f"{i} {j}"]
                return self.X_test[k], f"Classifier label: {int(self.encoded_test[k][2])}"

            # generate the nD data point on the fly using the inverse projection
            point = self.dbm_model.neural_network.decode([(i/self.img.shape[0], j/self.img.shape[1])])[0]
            return point, None

        def onclick(event):
            """ Open the data point in a new window when clicked on a pixel in training or testing set based on mouse position."""
            if event.inaxes == None:
                return

            if self.update_labels_by_circle_select:
                onclick_circle_strategy(event)
                return

            self.console.log("Clicked on: " + str(event.xdata) + ", " + str(event.ydata))
            j, i = int(event.xdata), int(event.ydata)

            if self.img[i][j] != TRAIN_DATA_POINT_MARKER:
                self.console.log("Data point not in training set")
                return

            # check if clicked on a data point that already was updated, then remove the update
            if f"{i} {j}" in self.expert_updates_labels_mapper:
                (_, point) = self.expert_updates_labels_mapper[f"{i} {j}"]
                point.remove()
                del self.expert_updates_labels_mapper[f"{i} {j}"]
                fig.canvas.draw_idle()
                self.updates_logger.log(f"Removed updates for point: ({j}, {i})")
                return

            # if clicked on a data point that was not updated, then add the update
            self.current_selected_point = ax.plot(j, i, 'bo', markersize=5)[0]

            # disable annotations on hover
            # if self.motion_event_cid is not None:
            #    fig.canvas.mpl_disconnect(self.motion_event_cid)
            # disable on click event
            if self.click_event_cid is not None:
                fig.canvas.mpl_disconnect(self.click_event_cid)
                self.click_event_cid = None

            # enable key press events
            self.key_event_cid = fig.canvas.mpl_connect('key_press_event', onkey)

            fig.canvas.draw_idle()

        def onkey(event):
            if self.current_selected_point is None:
                return

            if event.key.isdigit():
                self.current_selected_point_assigned_label = int(event.key)
                return

            (x, y) = self.current_selected_point.get_data()

            if event.key == 'escape' or event.key == 'enter':
                self.current_selected_point.remove()

                if event.key == 'escape':
                    self.updates_logger.log("Cancelled point move...")

                elif event.key == 'enter':
                    # showing the fix of the position
                    self.current_selected_point = ax.plot(x, y, 'b^')[0]
                    if self.current_selected_point_assigned_label is not None:
                        self.updates_logger.log(f"Assigned label: {self.current_selected_point_assigned_label} to point: ({x[0]}, {y[0]})")
                        self.expert_updates_labels_mapper[f"{y[0]} {x[0]}"] = (self.current_selected_point_assigned_label, self.current_selected_point)

                self.current_selected_point = None
                self.current_selected_point_assigned_label = None

                self.motion_event_cid = fig.canvas.mpl_connect('motion_notify_event', display_annotation)
                self.click_event_cid = fig.canvas.mpl_connect('button_press_event', onclick)
                fig.canvas.mpl_disconnect(self.key_event_cid)
                fig.canvas.draw_idle()
                return

        def onclick_circle_strategy(event):
            self.console.log("Clicked on: " + str(event.xdata) + ", " + str(event.ydata))
            x, y = int(event.xdata), int(event.ydata)
            self.current_selected_point = (x, y)
            self.release_event_cid = fig.canvas.mpl_connect('button_release_event', onrelease_circle_strategy)

        def onrelease_circle_strategy(event):
            if event.inaxes == None:
                return
            self.console.log("Released on: " + str(event.xdata) + ", " + str(event.ydata))
            (x0, y0) = self.current_selected_point
            x1, y1 = int(event.xdata), int(event.ydata)
            (cx, cy) = (x0 + (x1 - x0)/2, y0 + (y1 - y0)/2)  # circle center
            r = sqrt(((x1 - x0)/2)**2 + ((y1 - y0)/2)**2)   # circle radius
            self.update_labels_circle = Circle((cx, cy), r, color='black', fill=False)
            ax.add_artist(self.update_labels_circle)
            self.key_event_cid = fig.canvas.mpl_connect('key_press_event', onkey_circle_strategy)
            if self.release_event_cid is not None:
                fig.canvas.mpl_disconnect(self.release_event_cid)
            fig.canvas.draw_idle()

        def onkey_circle_strategy(event):
            if self.update_labels_circle is None:
                return

            if event.key.isdigit():
                self.current_selected_point_assigned_label = int(event.key)
                return

            (x, y) = self.update_labels_circle.get_center()
            r = self.update_labels_circle.get_radius()
            if event.key == "enter" and self.current_selected_point_assigned_label is not None:
                self.updates_logger.log(f"Assigned label: {self.current_selected_point_assigned_label} to circle: center ({x}, {y}), radius ({r})")
                positions = find_points_in_circle((x, y), r)
                for pos in positions:
                    point = ax.plot(pos[0], pos[1], 'b^')[0]
                    self.expert_updates_labels_mapper[f"{pos[1]} {pos[0]}"] = (self.current_selected_point_assigned_label, point)

            if event.key == "backspace":
                self.updates_logger.log(f"Removing changes in circle: center ({x},{y}), radius ({r})")
                positions = find_points_in_circle((x, y), r)
                for pos in positions:
                    if f"{pos[1]} {pos[0]}" in self.expert_updates_labels_mapper:
                        self.expert_updates_labels_mapper[f"{pos[1]} {pos[0]}"][1].remove()
                        del self.expert_updates_labels_mapper[f"{pos[1]} {pos[0]}"]

            self.update_labels_circle.remove()
            self.update_labels_circle = None
            fig.canvas.mpl_disconnect(self.key_event_cid)
            fig.canvas.draw_idle()

        def find_points_in_circle(circle_center, circle_radius):
            (cx, cy) = circle_center
            initial_x, initial_y = int(cx - circle_radius), int(cy - circle_radius)
            final_x, final_y = int(cx + circle_radius) + 1, int(cy + circle_radius) + 1
            initial_x = 0 if initial_x < 0 else initial_x
            initial_y = 0 if initial_y < 0 else initial_y
            final_x = self.img.shape[0] if final_x > self.img.shape[0] else final_x
            final_y = self.img.shape[0] if final_y > self.img.shape[0] else final_y

            positions = []
            for x in range(initial_x, final_x):
                for y in range(initial_y, final_y):
                    if (self.img[y, x] == TRAIN_DATA_POINT_MARKER) and ((x - cx)**2 + (y - cy)**2 <= circle_radius**2):
                        positions.append((x, y))
            return positions

        self.motion_event_cid = fig.canvas.mpl_connect('motion_notify_event', display_annotation)
        if connect_click_event:
            self.click_event_cid = fig.canvas.mpl_connect('button_press_event', onclick)



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import numpy as np

from ..AbstractNN import AbstractNN, SEED
from ...Logger import LoggerInterface, LoggerModel

DECODER_NAME = "decoder"
ENCODER_NAME = "encoder"
SSNP_NAME = "ssnp"
CLASSIFIER_NAME = "classifier"

DECODER_LOSS = "mean_squared_error"
CLASSIFIER_LOSS = "sparse_categorical_crossentropy"
DECODER_LOSS_WEIGHT = 1.0
CLASSIFIER_LOSS_WEIGHT = 0.125


class SSNP(AbstractNN):
    def __init__(self,
                 folder_path: str,
                 logger: LoggerInterface | None = None):
        """
            Creates an SSNP model.

            Args:
                folder_path: The path to the folder where the model will be saved/loaded.
                logger: The logger used to log the model's progress.
        """
        super().__init__(folder_path=folder_path, logger=logger, nn_name=SSNP_NAME)

    def __build__(self, input_shape: tuple = (28, 28), num_classes: int = 10, show_summary: bool = False):
        """
            Assembles the autoencoder model and compiles it. 

            Args:
                input_shape: The input and output shape of the autoencoder.
                show_summary: If True, the model summary will be printed.
        """

        output_size = 1
        for i in range(len(input_shape)):
            output_size *= input_shape[i]


        encoder = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(512, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  kernel_regularizer=tf.keras.regularizers.l2(0.0002)),
            tf.keras.layers.Dense(128, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(64, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(32, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(
                2, activation='sigmoid', bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
        ], name=ENCODER_NAME)

        decoder = tf.keras.models.Sequential([
            tf.keras.layers.Dense(32, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  kernel_regularizer=tf.keras.regularizers.l2(0.0002)),
            tf.keras.layers.Dense(64, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(128, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(512, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(output_size, activation='sigmoid',
                                  kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED)),  # type: ignore
            tf.keras.layers.Reshape(input_shape)
        ], name=DECODER_NAME)

        classifier = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(num_classes, activation=tf.nn.softmax,
                                  kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED))  # type: ignore
        ], name=CLASSIFIER_NAME)

        input_layer = tf.keras.Input(shape=input_shape, name="input")

        encoder_output = encoder(input_layer)
        decoder_output = decoder(encoder_output)
        classifier_output = classifier(decoder_output)

        self.neural_network = tf.keras.models.Model(inputs=input_layer,
                                                    outputs=[decoder_output, classifier_output],
                                                    name=SSNP_NAME)

        self.neural_network.compile(optimizer=tf.keras.optimizers.Adam(),
                                    loss={DECODER_NAME: DECODER_LOSS,
                                          CLASSIFIER_NAME: CLASSIFIER_LOSS},
                                    loss_weights={DECODER_NAME: DECODER_LOSS_WEIGHT,
                                                  CLASSIFIER_NAME: CLASSIFIER_LOSS_WEIGHT},
                                    metrics={DECODER_NAME: "accuracy", CLASSIFIER_NAME: "accuracy"})

        if show_summary:
            self.neural_network.summary(print_fn=self.console.log)

    def fit(self, X: np.ndarray, Y: np.ndarray,
            epochs: int = 100, batch_size: int = 128):
        """ Fits the model to the specified data.

        Args:
            X (np.ndarray): Train input values
            Y (np.ndarray): Train target values
            epochs (int, optional): The number of epochs. Defaults to 10.
            batch_size (int, optional): Data points used for one batch. Defaults to 128.
        """
        if self.neural_network is not None:
            self.console.log("Model already loaded. Skipping build.")
            self.encoder = self.neural_network.get_layer(ENCODER_NAME)
            self.decoder = self.neural_network.get_layer(DECODER_NAME)
            return

        self.console.log("Building model according to the data shape.")
        num_classes = len(np.unique(Y))
        self.__build__(input_shape=X.shape[1:], num_classes=num_classes, show_summary=True)

        stopping_callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', mode='min', patience=20, restore_best_weights=True)
        logger_callback = LoggerModel(name=SSNP_NAME, show_init=False, epochs=epochs, print_fn=self.console.log)

        self.console.log("Fitting model...")

        history = self.neural_network.fit(X, [X, Y],
                                          epochs=epochs,
                                          batch_size=batch_size,
                                          shuffle=True,
                                          validation_split=0.2,
                                          callbacks=[stopping_callback, logger_callback],
                                          verbose=0)

        self.console.log("Model fitted!")
        self.save(history)
        self.encoder = self.neural_network.get_layer(ENCODER_NAME)
        self.decoder = self.neural_network.get_layer(DECODER_NAME)

        # self.show_predictions(dataNd=x_test, labels=y_test)

    def encode(self, data: np.ndarray, verbose: int = 0):
        """ 
        Encodes the data using the encoder part of the autoencoder.

        Args:
            data (np.ndarray): The data to encode.
            verbose (int, optional): Verbosity level. Defaults to 0.

        Returns:
            np.ndarray: The encoded 2D data.
        """
        return self.encoder.predict(data, verbose=verbose)

    def decode(self, data: np.ndarray, verbose: int = 0):
        """ 
        sDecodes the data using the decoder part of the autoencoder.

        Args:
            data (np.ndarray): The data to decode.
            verbose (int, optional): Verbosity level. Defaults to 0.

        Returns:
            np.ndarray: The decoded nD data.
        """
        xNd = self.decoder.predict(data, verbose=verbose)
        return xNd



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import matplotlib.pyplot as plt
import PySimpleGUI as sg
import numpy as np
from PIL import Image, ImageTk

from ..Logger import Logger, LoggerGUI
from .GUIController import DBM_TECHNIQUES, PROJECTION_TECHNIQUES, DBM_NNINV_TECHNIQUE, CUSTOM_PROJECTION_TECHNIQUE, GUIController
from ..utils import BLACK_COLOR, WHITE_COLOR, RIGHTS_MESSAGE_1, RIGHTS_MESSAGE_2, APP_PRIMARY_COLOR, APP_FONT, APP_FONT_BOLD
from .utils import Collapsible

sg.theme('DarkBlue1')
TITLE = "Classifiers visualization tool"
WINDOW_SIZE = (1150, 700)
APP_ICON_PATH = os.path.join(os.path.dirname(__file__), "assets", "main_icon_b64.txt")
class GUI:
    def __init__(self):
        self.window = self.build_window()
        self.gui_logger = LoggerGUI(name="DBM logger", output=self.window["-LOGGER-"], update_callback=self.window.refresh)
        self.logger = Logger(name="GUI")
        self.controller = GUIController(self.window, self, self.gui_logger)

        # --------------- DBM ---------------
        self.dbm_plotter_gui = None

    def build_window(self):
        window = sg.Window(TITLE,
                           layout=self._get_layout(),
                           size=WINDOW_SIZE,
                           resizable=False,
                           )
        
        with open(APP_ICON_PATH, "rb") as f:
            iconb64 = f.read()

        window.finalize()
        window.set_icon(pngbase64=iconb64)
        
        return window

    def start(self):
        while True:
            event, values = self.window.read()

            if event == "Exit" or event == sg.WIN_CLOSED:
                break

            self.handle_event(event, values)

        self.stop()

    def stop(self):
        self.controller.stop()
        self.logger.log("Closing the application...")
        self.window.close()

    def _get_layout(self):
        data_files_list_column = [
            [
                sg.Button("Upload MNIST Data set", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, key="-UPLOAD MNIST DATA BTN-"),
            ],
            [
                sg.Button("Upload FASHION MNIST Data set", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, key="-UPLOAD FASHION MNIST DATA BTN-"),
            ],
            [
                sg.Button("Upload CIFAR10 Data set", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, key="-UPLOAD CIFAR10 DATA BTN-"),
            ],
            [ sg.HSeparator() ],
            [
                Collapsible([
                    [
                        sg.Text(text="Choose the data folder: ", font=APP_FONT),
                        sg.In(enable_events=True, key="-DATA FOLDER-", visible=False),
                        sg.Button("Browse folder", 
                          button_type=sg.BUTTON_TYPE_BROWSE_FOLDER, 
                          target=(sg.ThisRow, -1), 
                          button_color=(WHITE_COLOR, APP_PRIMARY_COLOR),
                          font=APP_FONT,
                          initial_folder=os.getcwd(),
                          expand_x=True,
                        ), 
                    ],
                    [
                        sg.Text("Choose the data file from the list: ", font=APP_FONT, expand_x=True),
                    ],
                    [
                       sg.pin(sg.Column([
                            [sg.Text(key="-DATA FOLDER TOUT-", font=APP_FONT, expand_x=True)],
                           ], key="-DATA FOLDER TOUT PIN-", visible=False, expand_x=True), 
                        shrink=True, expand_x=True), 
                    ],
                    [
                       sg.pin(sg.Column([
                            [sg.Text(key="-DATA FILE TOUT-", font=APP_FONT, expand_x=True)],
                           ], key="-DATA FILE TOUT PIN-", visible=False, expand_x=True), 
                        shrink=True, expand_x=True), 
                    ],
                    [
                        sg.Listbox(
                            values=[], enable_events=True, key="-DATA FILE LIST-", 
                            background_color=WHITE_COLOR, text_color=BLACK_COLOR, sbar_background_color=APP_PRIMARY_COLOR,
                            expand_x=True, size=(10, 10), font=APP_FONT_BOLD,
                        )
                    ],
                    [
                        sg.Button("Upload train data for DBM", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, key="-UPLOAD TRAIN DATA BTN-"),
                        sg.Button("Upload test data for DBM", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, key="-UPLOAD TEST DATA BTN-"),
                    ],
                    [
                        sg.Button("Upload Data set as folder", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, key="-UPLOAD FOLDER DATASET BTN-"),
                    ],
            
                ], "-UPLOAD DATA COLLAPSABLE-", "Upload the dataset from a folder", collapsed=True, visible=True), 
            ],
            [ sg.HSeparator() ],
            [
                sg.Text("Training data file: ", font=APP_FONT, key="-TRAIN DATA FILE-",  expand_x=True),
            ],
            [
                sg.Text("Training data shape: ", font=APP_FONT, key="-TRAIN DATA SHAPE-", expand_x=True),
            ],
            [
                sg.Text("Testing data file: ", font=APP_FONT, key="-TEST DATA FILE-", expand_x=True),
            ],
            [
                sg.Text("Testing data shape: ", font=APP_FONT, key="-TEST DATA SHAPE-", expand_x=True),
            ],
            [ sg.HSeparator()],
            [
                sg.Text(text="Choose the classifier folder", font=APP_FONT),
                sg.In(enable_events=True, key="-CLASSIFIER FOLDER-", visible=False),
                sg.Button("Browse folder", 
                          button_type=sg.BUTTON_TYPE_BROWSE_FOLDER, 
                          target=(sg.ThisRow, -1), 
                          button_color=(WHITE_COLOR, APP_PRIMARY_COLOR),
                          font=APP_FONT,
                          initial_folder=os.getcwd(),
                          expand_x=True,
                    ),
            ],
            [
                sg.Text("Choose the classifier file(.h5) or folder from the list: ", font=APP_FONT, expand_x=True),
            ],
            [
                sg.pin(sg.Column([
                            [sg.Text(key="-CLASSIFIER PATH TOUT-", font=APP_FONT, expand_x=True)],
                           ], key="-CLASSIFIER PATH TOUT PIN-", visible=False, expand_x=True), 
                shrink=True, expand_x=True), 
            ],
            [
                sg.Listbox(
                    values=[], enable_events=True, key="-CLASSIFIER FILE LIST-", 
                    background_color=WHITE_COLOR, text_color=BLACK_COLOR, sbar_background_color=APP_PRIMARY_COLOR,
                    expand_x=True, expand_y=True, font=APP_FONT_BOLD,
                )
            ],
            [
                sg.Button("Upload classifier", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, visible=True, key="-UPLOAD CLASSIFIER-"),
            ],
            [ sg.HSeparator()],
            [
                sg.Text("", expand_x=True)
            ],
            [
                sg.Text(RIGHTS_MESSAGE_1, font=APP_FONT, expand_x=True),
            ],
            [
                sg.Text(RIGHTS_MESSAGE_2, font=APP_FONT, expand_x=True),
            ]
        ]

        results_column = [
            [
                sg.Text("Which dbm technique would you like to use?", font=APP_FONT, size=(45, 1)),
                sg.Combo(
                    values=list(DBM_TECHNIQUES.keys()),
                    default_value=list(DBM_TECHNIQUES.keys())[0],
                    expand_x=True,
                    font=APP_FONT,
                    button_background_color=APP_PRIMARY_COLOR,
                    key="-DBM TECHNIQUE-",
                    enable_events=True,
                    background_color=WHITE_COLOR, text_color=BLACK_COLOR,
                ),
            ],
            [
                sg.pin(sg.Column([
                    [
                    sg.Text("Which Projection technique would you like to use?", font=APP_FONT, size=(45, 1), key="-PROJECTION TECHNIQUE TEXT-"),
                    sg.Combo(
                        values=PROJECTION_TECHNIQUES,
                        default_value=PROJECTION_TECHNIQUES[0],
                        expand_x=True,
                        font=APP_FONT,
                        button_background_color=APP_PRIMARY_COLOR,
                        key="-PROJECTION TECHNIQUE-",
                        enable_events=True,
                        background_color=WHITE_COLOR, text_color=BLACK_COLOR,
                    ),
                    ]
                ], expand_x=True, key="-PROJECTION TECHNIQUE COLUMN-", visible=False), shrink=True, expand_x=True)    
            ],
            [
                sg.pin(sg.Column([
                    [
                     Collapsible([
                            [
                                sg.Text(text="Choose the 2D data folder", key="-DATA 2D FOLDER TEXT-", font=APP_FONT),
                                sg.In(enable_events=True, key="-DATA 2D FOLDER-", visible=False),
                                sg.Button("Browse folder", 
                                    button_type=sg.BUTTON_TYPE_BROWSE_FOLDER, 
                                    target=(sg.ThisRow, -1), 
                                    button_color=(WHITE_COLOR, APP_PRIMARY_COLOR),
                                    font=APP_FONT,
                                    initial_folder=os.getcwd(),
                                    expand_x=True,
                                ),
                            ],
                            [
                                sg.Listbox(
                                    values=[], enable_events=True, key="-DATA 2D FILE LIST-", 
                                    background_color=WHITE_COLOR, text_color=BLACK_COLOR, sbar_background_color=APP_PRIMARY_COLOR,
                                    expand_x=True, size=(10, 10), font=APP_FONT_BOLD,
                                )
                            ],   
                            [
                                sg.Button("Upload 2D train data", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, expand_y=False, key="-UPLOAD 2D TRAIN DATA BTN-"),
                                sg.Button("Upload 2D test data", button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), font=APP_FONT, expand_x=True, expand_y=False, key="-UPLOAD 2D TEST DATA BTN-"),
                            ],
                        ], "-DATA 2D COLLAPSABLE-", "Select the files with 2D representation of the data", collapsed=True, visible=True),
                    ]
                ], key="-DATA 2D COLUMN-", visible=False, expand_x=True),
                shrink = True, expand_x = True)
            ],
            [
                sg.Text("Show Decision Boundary Mapper NN history: ", font=APP_FONT, expand_x=True, key="-DBM HISTORY TEXT-", visible=True),
                sg.Checkbox("", default=False, font=APP_FONT, key="-DBM HISTORY CHECKBOX-", visible=True),
            ],
            # [
            #    sg.Text("Image resolution of the Decision Boundary Mapper: ", size=(45,1), expand_x=True, key="-DBM IMAGE RESOLUTION TEXT-", visible=True),
            #    sg.InputText("256", key="-DBM IMAGE RESOLUTION INPUT-", visible=True, background_color=WHITE_COLOR, text_color=BLACK_COLOR),
            # ],
            [
                sg.Button("Show the Decision Boundary Mapping", font=APP_FONT, button_color=(WHITE_COLOR, APP_PRIMARY_COLOR), expand_x=True, visible=False, key="-DBM BTN-"),
            ],
            # ---------------------------------------------------------------------------------------------------
            [ sg.pin(sg.Column([
                [sg.Text("Decision boundary map: ", font=APP_FONT, expand_x=True, justification='center')],
                ], key="-DBM TEXT-", visible=False, expand_x=True, expand_y=False), 
                shrink=True, expand_x=True),
            ],
            [
                sg.pin(sg.Column([
                      [sg.Text("Loading... ", font=APP_FONT, expand_x=True, justification='center')],
                    ], key="-DBM IMAGE LOADING-", visible=False, expand_x=True, expand_y=False), 
                    shrink=True, expand_x=True),
            ],
            [sg.Image(key="-DBM IMAGE-", visible=False, expand_x=True, expand_y=False, enable_events=True)],
            [
                sg.HSeparator(),
            ],
            [
                sg.Column([
                    [sg.Text("Logger: ", font=APP_FONT,)],
                    [sg.Multiline("", font=APP_FONT, 
                                  expand_x=True, expand_y=True, key="-LOGGER-", 
                                  background_color=WHITE_COLOR, 
                                  text_color=BLACK_COLOR,
                                  sbar_background_color=APP_PRIMARY_COLOR,   
                                  auto_size_text=True)],
                ], expand_x=True, expand_y=True, key="-LOGGER COLUMN-"),
            ]
        ]

        layout = [
            [
                sg.Column(data_files_list_column, expand_x=True, expand_y=True),
                sg.VSeparator(),
                sg.Column(results_column, expand_x=True, expand_y=True),
            ],
        ]

        return layout

    def switch_visibility(self, elements, visible):
        for x in elements:
            self.window[x].update(visible=visible)
        self.window.refresh()

    def handle_event(self, event, values):
        EVENTS = {
            "-DATA FOLDER-": self.handle_select_data_folder_event,
            "-DATA 2D FOLDER-": self.handle_select_2d_data_folder_event,
            "-CLASSIFIER FOLDER-": self.handle_select_classifier_folder_event,
        
            "-DATA FILE LIST-": self.controller.handle_file_list_event,
            "-DATA 2D FILE LIST-": self.handle_2d_file_list_event,
          
            
            "-CLASSIFIER FILE LIST-": self.controller.handle_classifier_file_list_event,
            
            "-DBM TECHNIQUE-": self.handle_dbm_technique_event,
            "-DBM BTN-": self.handle_get_decision_boundary_mapping_event,
            "-DBM IMAGE-": self.handle_dbm_image_event,
            "-PROJECTION TECHNIQUE-": self.handle_projection_technique_event,
            
            "-UPLOAD 2D TRAIN DATA BTN-": self.controller.handle_upload_2d_data_event,
            "-UPLOAD 2D TEST DATA BTN-": self.controller.handle_upload_2d_data_event,
            "-UPLOAD TRAIN DATA BTN-": self.controller.handle_upload_train_data_event,
            "-UPLOAD TEST DATA BTN-": self.controller.handle_upload_test_data_event,
            "-UPLOAD MNIST DATA BTN-": self.controller.handle_upload_known_data_event,
            "-UPLOAD FASHION MNIST DATA BTN-": self.controller.handle_upload_known_data_event,
            "-UPLOAD CIFAR10 DATA BTN-": self.controller.handle_upload_known_data_event,
            "-UPLOAD FOLDER DATASET BTN-": self.controller.handle_upload_folder_dataset_event,
            
            "-UPLOAD CLASSIFIER-": self.controller.handle_upload_classifier_event,
            
            # collapsable components
            "-DATA 2D COLLAPSABLE-/-BUTTON-": self.handle_collapse_event,
            "-DATA 2D COLLAPSABLE-/-TITLE-": self.handle_collapse_event,
            "-UPLOAD DATA COLLAPSABLE-/-BUTTON-": self.handle_collapse_event,
            "-UPLOAD DATA COLLAPSABLE-/-TITLE-": self.handle_collapse_event,
        }

        EVENTS[event](event, values)
    
    def handle_collapse_event(self, event, values):
        collapsable_key = event.split('/')[0]
        component = self.window[collapsable_key]
        btn_component = self.window[collapsable_key + "/" +"-BUTTON-"]
        component.update(visible=not component.visible)
        btn_component.update(component.metadata[0] if component.visible else component.metadata[1])

    def handle_2d_file_list_event(self, event, values):
        filename = os.path.join(values["-DATA 2D FOLDER-"], values["-DATA 2D FILE LIST-"][0])
        self.gui_logger.log(f"Selected file: {filename}")
        
    def handle_select_data_folder_event(self, event, values):
        folder = values["-DATA FOLDER-"]
        self.window["-DATA FOLDER TOUT-"].update(folder)
        self.switch_visibility(["-DATA FOLDER TOUT PIN-"], True)
        files = self.controller.handle_select_data_folder(folder)
        self.window["-DATA FILE LIST-"].update(files)
        
    def handle_select_2d_data_folder_event(self, event, values):
        folder = values["-DATA 2D FOLDER-"]
        files = self.controller.handle_select_data_folder(folder)
        self.window["-DATA 2D FILE LIST-"].update(files)
        
    def handle_select_classifier_folder_event(self, event, values):
        folder = values["-CLASSIFIER FOLDER-"]
        f = self.controller.handle_select_classifier_folder(folder)
        self.window["-CLASSIFIER FILE LIST-"].update(f)
    
    def handle_dbm_technique_event(self, event, values):
        dbm_technique = values["-DBM TECHNIQUE-"]
        self.logger.log(f"DBM technique: {dbm_technique}")
        
        if dbm_technique == DBM_NNINV_TECHNIQUE:
            self.switch_visibility(["-PROJECTION TECHNIQUE COLUMN-"], True)
            return
        
        self.switch_visibility(["-PROJECTION TECHNIQUE COLUMN-", "-DATA 2D COLUMN-"], False)
        
    def handle_projection_technique_event(self, event, values):
        projection_technique = values["-PROJECTION TECHNIQUE-"]
        self.logger.log(f"Projection technique: {projection_technique}")
       
        if projection_technique == CUSTOM_PROJECTION_TECHNIQUE:
            self.switch_visibility(["-DATA 2D COLUMN-"], True)
            return
        
        self.switch_visibility(["-DATA 2D COLUMN-"], False)   
        
    def handle_changes_in_dbm_plotter(self):
        # ---------------------------------
        # update the dbm image
        img = Image.fromarray(np.uint8(self.dbm_plotter_gui.color_img*255))
        WINDOW_IMAGE_RESOLUTION = 256
        img.thumbnail((WINDOW_IMAGE_RESOLUTION, WINDOW_IMAGE_RESOLUTION), Image.ANTIALIAS)
        # Convert im to ImageTk.PhotoImage after window finalized
        image = ImageTk.PhotoImage(image=img)
        self.window["-DBM IMAGE-"].update(data=image)

        self.switch_visibility(["-DBM IMAGE LOADING-"], False)
        
        self.switch_visibility(["-DBM IMAGE-"], True)
        

    def show_dbm_history(self, training_history):
        # this is for plotting the training history
        plt.close()
        fig, [ax1, ax2] = plt.subplots(1, 2, figsize=(10, 5))
        ax1.set_title("Loss")
        ax1.plot(training_history["loss"], label="loss")
        ax1.plot(training_history["val_loss"], label="val_loss")
        ax1.legend()

        ax2.set_title("Decoder Accuracy")
        if "decoder_accuracy" in training_history and "val_decoder_accuracy" in training_history:
            ax2.plot(training_history["decoder_accuracy"], label="decoder_accuracy")
            ax2.plot(training_history["val_decoder_accuracy"], label="val_decoder_accuracy")
        else:
            ax2.plot(training_history["accuracy"], label="accuracy")
            ax2.plot(training_history["val_accuracy"], label="val_accuracy")
            
        ax2.legend()

        plt.show()

    def handle_get_decision_boundary_mapping_event(self, event, values):
        try:
            self.dbm_plotter_gui = self.controller.handle_get_decision_boundary_mapping_event(event, values)
            self.handle_changes_in_dbm_plotter()
        except Exception as e:
            self.gui_logger.error(e)
            # update loading state
            self.switch_visibility(["-DBM TEXT-", "-DBM IMAGE LOADING-","-DBM IMAGE-"], False)

        
    def handle_dbm_image_event(self, event, values):
        self.logger.log("Clicked on the dbm image")
        if self.dbm_plotter_gui is None:
            self.logger.warn("Nothing to show")
            return
        self.dbm_plotter_gui.start()

    def set_data_shape(self, data_shape: tuple):
        """This method sets the data shape of the data to be used for the DBM
        If the user imports data from a file (e.g. csv file), the data shape is unknown.
        This method is used to set the data shape after the data is imported, so that the data can be reshaped into the correct shape.
        Args:
            data_shape (tuple): The shape of the data
        """
        self.controller.data_shape = data_shape


# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt
import os

from ..DBM import DBM, FAST_DBM_STRATEGIES
from ..GUI import DBMPlotterGUI

from .utils import *


def DBM_usage_example():
    # import the dataset
    X_train, X_test, _, _ = import_data()

    # import the classifier
    classifier = import_classifier()

    # create the DBM
    dbm = DBM(classifier=classifier)

    X2d_train, X2d_test = import_2d_data()

    # use the DBM to get the decision boundary map, if you don't have the 2D projection of the data
    # the DBM will get it for you, you just need to specify the projection method you would like to use (t-SNE, PCA or UMAP)
    img, img_confidence, _, _ = dbm.generate_boundary_map(X_train,
                                                          X_test,
                                                          X2d_train=X2d_train,
                                                          X2d_test=X2d_test,
                                                          resolution=256,
                                                          load_folder=os.path.join("tmp", "MNIST", "DBM"),
                                                          fast_decoding_strategy=FAST_DBM_STRATEGIES.NONE,
                                                          )

    # if you have the 2D projection of the data, you can use the following function to get the decision boundary map
    """
    X2d_train, X2d_test = None, None # get the 2D projection of the data by yourself
    img, img_confidence, _, _ = dbm.generate_boundary_map(X_train,
                                                          X_test,
                                                          X2d_train=X2d_train, 
                                                          X2d_test=X2d_test,
                                                          load_folder=os.path.join("tmp", "MNIST", "DBM"), 
                                                          resolution=256)
    """

    # make the decision boundary map pretty, by adding the colors and the confidence
    COLORS_MAPPER = {
        -2: [0, 0, 0],  # setting original test data to black
        -1: [1, 1, 1],  # setting original train data to white
        0: [1, 0, 0],
        1: [0, 1, 0],
        2: [0, 0, 1],
        3: [1, 1, 0],
        4: [0, 1, 1],
        5: [1, 0, 1],
        6: [0.5, 0.5, 0.5],
        7: [0.5, 0, 0],
        8: [0, 0.5, 0],
        9: [0, 0, 0.5]
    }

    color_img = np.zeros((img.shape[0], img.shape[1], 4))
    for i, j in np.ndindex(img.shape):
        color_img[i, j] = COLORS_MAPPER[img[i, j]] + [img_confidence[i, j]]

    # plot the decision boundary map
    plt.title("Decision boundary map")
    plt.axis("off")
    plt.imshow(color_img)
    plt.show()

    # use the dbm to get the inverse projection errors
    img_inverse_projection_errors = dbm.generate_inverse_projection_errors(resolution=256)
    # plot the inverse projection errors
    fig, ax = plt.subplots()
    ax.set_title("Inverse projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_inverse_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()

    # use the dbm to get the projection errors
    img_projection_errors = dbm.generate_projection_errors()
    # plot the projection errors
    fig, ax = plt.subplots()
    ax.set_title("Projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()


def DBM_usage_example_GUI():
    # import the dataset
    X_train, X_test, Y_train, Y_test = import_data()

    # import the classifier
    classifier = import_classifier()

    # import the 2D projection of the data
    X2d_train, X2d_test = import_2d_data()

    # create the DBM
    dbm = DBM(classifier=classifier)

    # use the DBM to get the decision boundary map, if you don't have the 2D projection of the data
    # the DBM will get it for you, you just need to specify the projection method you would like to use (t-SNE, PCA or UMAP)
    img, img_confidence, encoded_training_data, encoded_testing_data = dbm.generate_boundary_map(X_train,
                                                                                                 X_test,
                                                                                                 X2d_train=X2d_train,
                                                                                                 X2d_test=X2d_test,
                                                                                                 resolution=256,
                                                                                                 load_folder=os.path.join("tmp", "MNIST", "DBM"),
                                                                                                 projection = "t-SNE",
                                                                                                 )

    
    dbm_plotter_gui = DBMPlotterGUI(dbm_model=dbm,
                                    img=img,
                                    img_confidence=img_confidence,
                                    encoded_train=encoded_training_data,
                                    encoded_test=encoded_testing_data,
                                    X_train=X_train,
                                    Y_train=Y_train,
                                    X_test=X_test,
                                    Y_test=Y_test,
                                    # this is the folder where the DBM will save the changes in data the user makes
                                    save_folder=os.path.join("tmp", "MNIST", "DBM"),
                                    projection_technique="t-SNE",
                                    )
    dbm_plotter_gui.start()



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from termcolor import colored
from tensorflow.keras.callbacks import Callback


class LoggerModel(Callback):
    def __init__(self, active: bool = True, name: str = "Neural Network", info_color: str = "magenta", show_init: bool = True, epochs: int = 100, print_fn=None):
        """ Initialize the logger

        Args:
            active (bool, optional): Defaults to True.
            name (str, optional): Defaults to "Logger".
        """
        self.active = active
        self.name = name
        self.info_color = info_color
        self.epochs = epochs
        self.print_fn = print_fn
        if show_init:
            sep = "=" * 30
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color))
            print(colored(f"[INFO] [{time}] [{self.name}] {self.name} initialized", self.info_color))
            print(colored(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color))

    def on_epoch_end(self, epoch, logs={}):
        if self.active:
            logs = ", ".join([f"{key}: {value:.4f}" for key, value in logs.items()])
            print(colored(f"[INFO] [{self.name}] [Epoch {epoch}/{self.epochs}] {logs}", self.info_color))
            if self.print_fn is not None:
                self.print_fn(f"[Epoch {epoch}/{self.epochs}] {logs}")


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import PySimpleGUI as sg
from ..utils import APP_FONT

def Collapsible(layout, key, title='', arrows=(sg.SYMBOL_UP, sg.SYMBOL_DOWN), collapsed=False, visible=True):
    """
    User Defined Element
    A "collapsable section" element. Like a container element that can be collapsed and brought back
    :param layout:Tuple[List[sg.Element]]: The layout for the section
    :param key:Any: Key used to make this section visible / invisible
    :param title:str: Title to show next to arrow
    :param arrows:Tuple[str, str]: The strings to use to show the section is (Open, Closed).
    :param collapsed:bool: If True, then the section begins in a collapsed state
    :return:sg.Column: Column including the arrows, title and the layout that is pinned
    """
    return sg.Column([
                      [sg.T((arrows[1] if collapsed else arrows[0]), enable_events=True, k=key+'/-BUTTON-', expand_x=True, font=APP_FONT),
                       sg.T(title, enable_events=True, key=key+'/-TITLE-', expand_x=True, font=APP_FONT)],
                      [sg.pin(
                            sg.Column(layout, key=key, visible=not collapsed, metadata=arrows, expand_x=True, expand_y=True), 
                          shrink=True, expand_x=True, expand_y=True)]], 
                     pad=(0,0), visible=visible, expand_x=True)



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import threading
import time
from playsound import playsound
from .. import LoggerInterface


def track_time_wrapper(logger: LoggerInterface | None = None):
    """ Decorator that tracks the time it takes for a function to run.

    Args:
        logger (LoggerInterface): The logger to use. If None, prints to stdout.
    """
    def function_wrapper(func):
        def wrapper(*args, **kwargs):
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            if logger is None:
                print(f"{func.__name__} took {end-start} seconds")
            else:
                logger.log(f"{func.__name__} took {end-start} seconds")
            return result
        return wrapper
    return function_wrapper

def generate_class_name_mapper(file: str):
    mapper = {}
    with open(file, "r") as f:
        lines = f.readlines()
    for line in lines:
        class_num, class_name = int(line.split(" ")[0]) - 1, line.split(" ")[1].replace("\n", "")
        mapper[class_num] = class_name
    
    def class_name_mapper(x: int):
        return mapper[x]
    return class_name_mapper

def get_latest_created_file_from_folder(folder_path: str) -> str:
    """
    Get the latest created file in a folder
    
    Args:
        folder_path (str): The folder in which to search for the latest created file
    
    Returns:
        latest_file_path (str): The path to the latest created file in the folder
    
    Throws:
        Exception if no file in the folder
    """
    # Get a list of all files in the folder
    files = os.listdir(folder_path)

    # Filter out subdirectories (if any) and get the latest created file
    latest_file = None
    latest_timestamp = 0

    for file in files:
        file_path = os.path.join(folder_path, file)

        # Check if it's a file (not a directory)
        if os.path.isfile(file_path):
            file_timestamp = os.path.getctime(file_path)

            # Compare the file's creation timestamp to the latest found
            if file_timestamp > latest_timestamp:
                latest_timestamp = file_timestamp
                latest_file = file_path

    if latest_file is None:
        raise Exception("Could not find latest file")

    return latest_file


TIMER_WARNING_INTERVAL = 10
TIMER_DURATION = 5


def run_timer(update_callback = lambda msg, time_up: print(msg), 
              stop_event: threading.Event = threading.Event(), 
              timer_duration: int = TIMER_DURATION, 
              timer_warning_interval: int = TIMER_WARNING_INTERVAL, 
              end_play_sound_path: str | None = None,
              indefinite_timer: bool = False
              ):
    """Runs a timer and plays a sound 5 times when time elapsed
    
    Args:
        update_callback (function: (message: str, time_up: bool) -> None, optional): the function to be executed once per second. Defaults to lambda msg: print(msg)
        stop_event (threading.Event, optional): stop timer event. Defaults to threading.Event()
        timer_duration (int, optional): The duration of the timer in seconds. Defaults to TIMER_DURATION
        timer_warning_interval (int, optional): The interval in seconds in which a sound is played after the expiration of the timer_duration
        end_play_sound_path (str, optional): The path to the sound file to play when time elapsed. Defaults to None
        indefinite_timer (bool, optional): The type of timer, if is set to False the time will stop after the timer_duration, else it will run until stop_event. Defaults to False
    WARNING:
        This function should run in a separate thread in order to not block the main thread
        
    Example:
        # Create a time stop event
        stop_timer_event = threading.Event()
        # Create a separate thread for the timer
        timer_thread = threading.Thread(target=run_timer, args=(update_callback=..., stop_event=stop_timer_event, timer_duration=60, timer_warning_interval=10, end_play_sound_path=..., indefinite_timer = True))
        # Start the timer thread
        timer_thread.start()

        ...
        
        # Stop the timer thread
        stop_timer_event.set()  

    """

    def play_sound(filename: str, amount=1):
        for _ in range(amount):
            playsound(filename)

    sound_thread = None
    start_time = time.time()

    while not stop_event.is_set():
        elapsed_time = time.time() - start_time
        remaining_time = int(timer_duration - elapsed_time)
        if (end_play_sound_path is not None) and ((remaining_time == 0) or (remaining_time < 0 and remaining_time % timer_warning_interval == 0)):
            if sound_thread is None:
                sound_thread = threading.Thread(target=play_sound, args=(end_play_sound_path, 5))
                sound_thread.start()
            elif not sound_thread.is_alive():
                sound_thread = threading.Thread(target=play_sound, args=(end_play_sound_path, 5))
                sound_thread.start()
        
        if not indefinite_timer and remaining_time <= 0:
            stop_event.set()

        minutes, seconds = divmod(abs(remaining_time), 60)
        time_up = remaining_time < 0
        sign = "-" if time_up else ""
        update_callback(f"Remaining Time: {sign}{minutes:02}:{seconds:02}", time_up)
        time.sleep(1)
        
    if sound_thread is not None:
        sound_thread.join()

    


# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from shutil import rmtree
import os

# remove the old dist folder
rmtree('dist', ignore_errors=True)
# remove the old documentation folder
rmtree('docs', ignore_errors=True)

os.system("pdoc --html src/decision_boundary_mapper -o docs")

setup(
    name='decision-boundary-mapper',
    version='0.6.7',
    license='MIT',
    author="Cristian Grosu",
    author_email='c.grosu@students.uu.nl',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/cristi2019255/MasterThesis2023',
    keywords='Decision Boundary Mapper',
    description='A tool for visualizing the decision boundary of a machine learning model.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=[
        'dask',
        'keras',
        'matplotlib',
        'numba',
        'numba_progress',
        'numpy',
        'opfython',
        'Pillow',
        'PySimpleGUI',
        'playsound',
        'pydot',
        'pandas',
        'scikit_learn',
        'scipy',
        'tensorflow',
        'termcolor',
        'tqdm',
        'umap',
        'umap-learn',
    ],
)



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt

from ..DBM import SDBM, NNArchitecture
from ..GUI import DBMPlotterGUI

from .utils import *


def SDBM_usage_example():
    # import the dataset
    X_train, X_test, Y_train, Y_test = import_data()

    # import the classifier
    classifier = import_classifier()

    # create the DBM
    sdbm = SDBM(classifier=classifier)

    # use the SDBM to get the decision boundary map
    img, img_confidence, _, _ = sdbm.generate_boundary_map(X_train, Y_train,
                                                           X_test, Y_test,
                                                           nn_architecture=NNArchitecture.SSNP,
                                                           load_folder=os.path.join("tmp", "MNIST", "SDBM"),
                                                            resolution=256)

    # make the decision boundary map pretty, by adding the colors and the confidence
    COLORS_MAPPER = {
        -2: [0, 0, 0],  # setting original test data to black
        -1: [1, 1, 1],  # setting original train data to white
        0: [1, 0, 0],
        1: [0, 1, 0],
        2: [0, 0, 1],
        3: [1, 1, 0],
        4: [0, 1, 1],
        5: [1, 0, 1],
        6: [0.5, 0.5, 0.5],
        7: [0.5, 0, 0],
        8: [0, 0.5, 0],
        9: [0, 0, 0.5]
    }

    color_img = np.zeros((img.shape[0], img.shape[1], 4))
    for i, j in np.ndindex(img.shape):
        color_img[i, j] = COLORS_MAPPER[img[i, j]] + [img_confidence[i, j]]

    # plot the decision boundary map
    plt.title("Decision boundary map")
    plt.axis("off")
    plt.imshow(color_img)
    plt.show()

    # use the SDBM to get the inverse projection errors
    img_inverse_projection_errors = sdbm.generate_inverse_projection_errors(
        resolution=256)
    # plot the inverse projection errors
    fig, ax = plt.subplots()
    ax.set_title("Inverse projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_inverse_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()

    # use the SDBM to get the projection errors
    img_projection_errors = sdbm.generate_projection_errors()
    # plot the projection errors
    fig, ax = plt.subplots()
    ax.set_title("Projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()


def SDBM_usage_example_GUI():
    # import the dataset
    X_train, X_test, Y_train, Y_test = import_data()

    # import the classifier
    classifier = import_classifier()

    # create the DBM
    sdbm = SDBM(classifier=classifier)

    # use the SDBM to get the decision boundary map
    img, img_confidence, encoded_training_data, encoded_testing_data = sdbm.generate_boundary_map(X_train, Y_train,
                                                                                                  X_test, Y_test,
                                                                                                  resolution=256,
                                                                                                  nn_architecture=NNArchitecture.SSNP,
                                                                                                  load_folder=os.path.join("tmp", "MNIST", "SDBM"),
                                                                                                  )

    dbm_plotter_gui = DBMPlotterGUI(dbm_model=sdbm,
                                    img=img,
                                    img_confidence=img_confidence,
                                    encoded_train=encoded_training_data,
                                    encoded_test=encoded_testing_data,
                                    X_train=X_train,
                                    Y_train=Y_train,
                                    X_test=X_test,
                                    Y_test=Y_test,
                                    save_folder=os.path.join("tmp", "MNIST", "SDBM"),
                                    )
    dbm_plotter_gui.start()
